from pathlib import Path
from unittest import TestCase

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestWarmingUpUnitCases(TestCase):
    def test_1a(self):
        import models.unit_cases.case_1a.src.run_1a as run_1a
        from models.unit_cases.case_1a.src.run_1a import HeatProblem

        base_folder = Path(run_1a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_optimization_problem(HeatProblem, base_folder=base_folder)

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())

    def test_2a(self):
        import models.unit_cases.case_2a.src.run_2a as run_2a
        from models.unit_cases.case_2a.src.run_2a import HeatProblem

        base_folder = Path(run_2a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_optimization_problem(HeatProblem, base_folder=base_folder)

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())

    def test_3a(self):
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblem

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_optimization_problem(HeatProblem, base_folder=base_folder)

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())



from pathlib import Path
from unittest import TestCase

import numpy as np


from rtctools.util import run_optimization_problem


class TestMILPGasSourceSink(TestCase):
    """Unit tests for the MILP test case of a source, a pipe, a sink"""

    def test_source_sink(self):
        import models.unit_cases_gas.source_sink.src.run_source_sink as example
        from models.unit_cases_gas.source_sink.src.run_source_sink import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        results = run_optimization_problem(GasProblem, base_folder=base_folder).extract_results()

        # Test if mass conserved
        np.testing.assert_allclose(
            results["GasProducer_0876.GasOut.Q"], results["GasDemand_a2d8.GasIn.Q"]
        )

        # Test if head is going down
        np.testing.assert_array_less(results["Pipe_4abc.GasOut.H"], results["Pipe_4abc.GasIn.H"])



import json
import xml.etree.ElementTree as ET  # noqa: N817
from typing import Tuple

import requests

from rtctools_heat_network._heat_loss_u_values_pipe import heat_loss_u_values_pipe


def main():
    """
    This script is meant to create a local update of the json file that has the standard pipe
    properties. These pipe classes are then used to select the most desireble pipe class

    Returns
    -------
    Nothing, it writes json file.
    """
    logstore = requests.get("https://edr.hesi.energy/api/edr_list?category=Assets/Logstore").json()
    asset_id_map = {x["title"].rsplit(".", maxsplit=1)[0]: x["key"] for x in logstore}

    # Build a dictionary of EDR pipe name to its properties.
    pipe_properties_map = {}

    for a, key in asset_id_map.items():
        xml_string = requests.get(f"https://edr.hesi.energy/api/edr_item?id={key}").json()[
            "esdl_string"
        ]
        tree = ET.fromstring(xml_string)

        inner_diameter = float(tree.get("innerDiameter"))

        investmet_cost = float((tree.findall(".//investmentCosts")[0]).get("value"))

        components = tree.findall(".//component")

        insulation_thicknesses = []
        conductivities_insulation = []

        for c in components:
            insulation_thicknesses.append(float(c.get("layerWidth")))
            matters = c.findall("matter")
            assert len(matters) == 1
            conductivities_insulation.append(float(matters[0].get("thermalConductivity")))

        u_1, u_2 = heat_loss_u_values_pipe(
            inner_diameter=inner_diameter,
            insulation_thicknesses=insulation_thicknesses,
            conductivities_insulation=conductivities_insulation,
        )

        pipe_properties_map[a] = {
            "inner_diameter": inner_diameter,
            "u_1": u_1,
            "u_2": u_2,
            "insulation_thicknesses": insulation_thicknesses,
            "conductivies_insulation": conductivities_insulation,
            "investment_costs": investmet_cost,
            "xml_string": xml_string,
        }

    # Sort the list based on prefix and diameter
    def _sort_series_dn(name: str) -> Tuple[str, int]:
        """
        This function sorts the list based on prefix and diameter
        Parameters
        ----------
        name : string with name of the pipe

        Returns
        -------
        Prefix of the name and the DN size in at int
        """
        a, b = name.rsplit("-", maxsplit=1)
        return a, int(b)

    pipe_properties_map = dict(
        sorted(pipe_properties_map.items(), key=lambda x: _sort_series_dn(x[0]))
    )

    # Export it to an indented (=human readable) JSON file
    json.dump(pipe_properties_map, open("_edr_pipes.json", "w"), indent=4)


if __name__ == "__main__":
    main()



import json
import locale
import logging
import os
import time
import xml.etree.ElementTree as ET  # noqa: N817
from pathlib import Path

from influxdb import InfluxDBClient

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import (
    CachingQPSol,
    SinglePassGoalProgrammingMixin,
)
from rtctools.util import run_optimization_problem

from rtctools_heat_network.esdl.esdl_mixin import ESDLMixin
from rtctools_heat_network.head_loss_mixin import HeadLossOption
from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.workflows.goals.minimize_tco_goal import MinimizeTCO
from rtctools_heat_network.workflows.io.write_output import ScenarioOutput
from rtctools_heat_network.workflows.utils.adapt_profiles import (
    adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day,
)
from rtctools_heat_network.workflows.utils.helpers import main_decorator


DB_HOST = "172.17.0.2"
DB_PORT = 8086
DB_NAME = "Warmtenetten"
DB_USER = "admin"
DB_PASSWORD = "admin"

logger = logging.getLogger("WarmingUP-MPC")
logger.setLevel(logging.INFO)

locale.setlocale(locale.LC_ALL, "")

ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}

WATT_TO_MEGA_WATT = 1.0e6
WATT_TO_KILO_WATT = 1.0e3


class TargetHeatGoal(Goal):
    priority = 1

    order = 2

    def __init__(self, state, target):
        self.state = state

        self.target_min = target
        self.target_max = target
        try:
            self.function_range = (-1.0e6, max(2.0 * max(target.values), 1.0e6))
            self.function_nominal = max(np.median(target.values), 1.0e6)
        except Exception:
            self.function_range = (-1.0e6, max(2.0 * target, 1.0e6))
            self.function_nominal = max(target, 1.0e6)

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class EndScenarioSizing(
    ScenarioOutput,
    HeatMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    Goal priorities are:
    1. minimize TCO = Capex + Opex*lifetime
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._override_hn_options = {}

        self._number_of_years = 30.0

        self.__indx_max_peak = None
        self.__day_steps = 5

        # self._override_pipe_classes = {}

        # variables for solver settings
        self._qpsol = None

        self._hot_start = False

        # Store (time taken, success, objective values, solver stats) per priority
        self._priorities_output = []
        self.__priority = None
        self.__priority_timer = None

        self.__heat_demand_bounds = dict()
        self.__heat_demand_nominal = dict()

    def _get_runinfo_path_root(self):
        runinfo_path = Path(self.esdl_run_info_path).resolve()
        tree = ET.parse(runinfo_path)
        return tree.getroot()

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters["peak_day_index"] = self.__indx_max_peak
        parameters["time_step_days"] = self.__day_steps
        parameters["number_of_years"] = self._number_of_years
        return parameters

    def pipe_classes(self, p):
        return self._override_pipe_classes.get(p, [])

    def pre(self):
        self._qpsol = CachingQPSol()

        super().pre()

        # parameters = self.parameters(0)
        # bounds = self.bounds()

        # TODO: these constraints do no longer work as we now have varying size timesteps
        # for s in self._setpoint_constraints_sources:
        #     # Here we enforce that over the full time horizon no setpoint changes can be done
        #     self._timed_setpoints[s] = (len(self.times()), 0)
        #
        # # Mixed-interger formulation of component setpoint
        # for component_name in self._timed_setpoints.keys():
        #     # Make 1 variable per component (so not per control
        #     # variable) which represents if the setpoint of the component
        #     # is changed (1) is not changed (0) in a timestep
        #     change_setpoint_var = f"{component_name}._change_setpoint_var"
        #     self._component_to_change_setpoint_map[component_name] = change_setpoint_var
        #     self._change_setpoint_var[change_setpoint_var] = ca.MX.sym(change_setpoint_var)
        #     self._change_setpoint_bounds[change_setpoint_var] = (0, 1.0)

    def read(self):
        """
        Reads the yearly profile with hourly time steps and adapt to a daily averaged profile
        except for the day with the peak demand.
        """
        super().read()

        (
            self.__indx_max_peak,
            self.__heat_demand_nominal,
        ) = adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day(self, self.__day_steps)

        logger.info("HeatProblem read")

    def bounds(self):
        bounds = super().bounds()
        bounds.update(self.__heat_demand_bounds)
        return bounds

    def variable_nominal(self, variable):
        try:
            return self.__heat_demand_nominal[variable]
        except KeyError:
            return super().variable_nominal(variable)

    def heat_network_options(self):
        # TODO: make empty placeholder in HeatProblem we don't know yet how to put the global
        #  constraints in the ESDL e.g. min max pressure
        options = super().heat_network_options()
        options["minimum_velocity"] = 0.001
        options["maximum_velocity"] = 3.0
        options["maximum_temperature_der"] = np.inf
        # options["neglect_pipe_heat_losses"] = True
        options["heat_loss_disconnected_pipe"] = True
        options["head_loss_option"] = HeadLossOption.NO_HEADLOSS
        # options.update(self._override_hn_options)
        return options

    def esdl_heat_model_options(self):
        """Overwrites the fraction of the minimum tank volume"""
        options = super().esdl_heat_model_options()
        options["min_fraction_tank_volume"] = 0.0
        return options

    def path_goals(self):
        goals = super().path_goals().copy()
        bounds = self.bounds()

        for demand in self.heat_network_components["demand"]:
            # target = self.get_timeseries(f"{demand}.target_heat_demand_peak")
            target = self.get_timeseries(f"{demand}.target_heat_demand")
            if bounds[f"{demand}.HeatIn.Heat"][1] < max(target.values):
                logger.warning(
                    f"{demand} has a flow limit, {bounds[f'{demand}.HeatIn.Heat'][1]}, "
                    f"lower that wat is required for the maximum demand {max(target.values)}"
                )
            state = f"{demand}.Heat_demand"

            goals.append(TargetHeatGoal(state, target))
        return goals

    def goals(self):
        goals = super().goals().copy()
        # We do a minization of TCO consisting of CAPEX and OPEX over 25 years
        # CAPEX is based upon the boolean placement variables and the optimized maximum sizes
        # Note that CAPEX for geothermal and ATES is also dependent on the amount of doublets
        # In practice this means that the CAPEX is mainly driven by the peak day problem
        # The OPEX is based on the Source strategy which is computed on the __daily_avg variables
        # The OPEX thus is based on an avg strategy and discrepancies due to fluctuations intra-day
        # are possible.
        # The idea behind the two timelines is that the optimizer can make the OPEX vs CAPEX
        # trade-offs

        goals.append(MinimizeTCO(priority=2, number_of_years=self._number_of_years))

        return goals

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        for a in self.heat_network_components.get("ates", []):
            stored_heat = self.state_vector(f"{a}.Stored_heat")
            constraints.append(((stored_heat[-1] - stored_heat[0]), 0.0, np.inf))

        for b in self.heat_network_components.get("buffer", {}):
            vars = self.state_vector(f"{b}.Heat_buffer")
            symbol_stored_heat = self.state_vector(f"{b}.Stored_heat")
            constraints.append((symbol_stored_heat[self.__indx_max_peak], 0.0, 0.0))
            for i in range(len(self.times())):
                if i < self.__indx_max_peak or i > (self.__indx_max_peak + 23):
                    constraints.append((vars[i], 0.0, 0.0))

        return constraints

    def history(self, ensemble_member):
        return AliasDict(self.alias_relation)

    def __state_vector_scaled(self, variable, ensemble_member):
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "gurobi"
        gurobi_options = options["gurobi"] = {}
        gurobi_options["MIPgap"] = 0.02
        gurobi_options["threads"] = 4
        gurobi_options["LPWarmStart"] = 2

        return options

    def solver_success(self, solver_stats, log_solver_failure_as_error):
        success, log_level = super().solver_success(solver_stats, log_solver_failure_as_error)

        # Allow time-outs for CPLEX and CBC
        if (
            solver_stats["return_status"] == "time limit exceeded"
            or solver_stats["return_status"] == "stopped - on maxnodes, maxsols, maxtime"
        ):
            if self.objective_value > 1e10:
                # Quick check on the objective value. If no solution was
                # found, this is typically something like 1E50.
                return success, log_level

            return True, logging.INFO
        else:
            return success, log_level

    def priority_started(self, priority):
        self.__priority = priority
        self.__priority_timer = time.time()

        super().priority_started(priority)

    def priority_completed(self, priority):
        super().priority_completed(priority)

        self._hot_start = True

        time_taken = time.time() - self.__priority_timer
        self._priorities_output.append(
            (
                priority,
                time_taken,
                True,
                self.objective_value,
                self.solver_stats,
            )
        )

    def post(self):
        # In case the solver fails, we do not get in priority_completed(). We
        # append this last priority's statistics here in post().
        # TODO: check if we still need this small part of code below
        success, _ = self.solver_success(self.solver_stats, False)
        if not success:
            time_taken = time.time() - self.__priority_timer
            self._priorities_output.append(
                (
                    self.__priority,
                    time_taken,
                    False,
                    self.objective_value,
                    self.solver_stats,
                )
            )

        super().post()
        results = self.extract_results()
        parameters = self.parameters(0)
        # bounds = self.bounds()
        # Optimized ESDL
        self._write_updated_esdl()

        for d in self.heat_network_components.get("demand", []):
            realized_demand = results[f"{d}.Heat_demand"]
            target = self.get_timeseries(f"{d}.target_heat_demand").values
            timesteps = np.diff(self.get_timeseries(f"{d}.target_heat_demand").times)
            parameters[f"{d}.target_heat_demand"] = target.tolist()
            delta_energy = np.sum((realized_demand - target)[1:] * timesteps / 1.0e9)
            if delta_energy >= 1.0:
                logger.warning(f"For demand {d} the target is not matched by {delta_energy} GJ")

        if self.esdl_run_info_path is None:
            logger.warning(
                "Not writing results since no esdl path to write the files to is specified"
            )
            return

        root = self._get_runinfo_path_root()
        parameters_dict = dict()
        workdir = root.findtext("pi:outputResultsFile", namespaces=ns)
        if workdir is None:
            logger.error("No workdir specified, skipping writing results")
            return

        parameter_path = os.path.join(workdir, "parameters.json")
        for key, value in parameters.items():
            new_value = value  # [x for x in value]
            parameters_dict[key] = new_value
        if parameter_path is None:
            workdir = root.findtext("pi:workDir", namespaces=ns)
            parameter_path = os.path.join(workdir, "parameters.json")
            if not Path(workdir).is_absolute():
                parameter_path = Path(workdir).resolve().parent
                parameter_path = os.path.join(parameter_path.__str__() + "parameters.json")
        with open(parameter_path, "w") as file:
            json.dump(parameters_dict, fp=file)

        # root = self._get_runinfo_path_root()
        # bounds_dict = dict()
        # bounds_path = root.findtext("pi:outputResultsFile", namespaces=ns)
        # bounds_path = os.path.join(workdir, "bounds.json")
        # for key, value in bounds.items():
        #     if "Stored_heat" not in key:
        #         new_value = value  # [x for x in value]
        #         # if len(new_value) == 1:
        #         #     new_value = new_value[0]
        #         bounds_dict[key] = new_value
        # if bounds_path is None:
        #     workdir = root.findtext("pi:workDir", namespaces=ns)
        #     bounds_path = os.path.join(workdir, "bounds.json")
        #     if not Path(workdir).is_absolute():
        #         bounds_path = Path(workdir).resolve().parent
        #         bounds_path = os.path.join(bounds_path.__str__() + "bounds.json")
        # with open(bounds_path, "w") as file:
        #     json.dump(bounds_dict, fp=file)

        root = self._get_runinfo_path_root()
        results_path = root.findtext("pi:outputResultsFile", namespaces=ns)
        results_dict = dict()

        for key, values in results.items():
            new_value = values.tolist()
            if len(new_value) == 1:
                new_value = new_value[0]
            results_dict[key] = new_value

        results_path = os.path.join(workdir, "results.json")
        if results_path is None:
            workdir = root.findtext("pi:workDir", namespaces=ns)
            results_path = os.path.join(workdir, "results.json")
            if not Path(workdir).is_absolute():
                results_path = Path(workdir).resolve().parent
                results_path = os.path.join(results_path.__str__() + "results.json")
        with open(results_path, "w") as file:
            json.dump(results_dict, fp=file)


def connect_database():
    client = InfluxDBClient(
        host=DB_HOST, port=DB_PORT, username=DB_USER, password=DB_PASSWORD, database=DB_NAME
    )
    if DB_NAME not in client.get_list_database():
        client.create_database(DB_NAME)
    return client


def format_datetime(dt):
    date, time = dt.split(" ")
    day, month, year = date.split("-")
    ndate = year + "-" + month + "-" + day
    ntime = time + ":00+0000"
    return ndate + "T" + ntime


class EndScenarioSizingHIGHS(EndScenarioSizing):
    def post(self):
        super().post()

        self._write_updated_esdl()

    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        highs_options["mip_rel_gap"] = 0.02

        options["gurobi"] = None

        return options


class EndScenarioSizingStaged(EndScenarioSizing):
    _stage = 0

    def __init__(self, stage=None, boolean_bounds=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._stage = stage
        self.__boolean_bounds = boolean_bounds

    def heat_network_options(self):
        options = super().heat_network_options()
        if self._stage == 1:
            options["neglect_pipe_heat_losses"] = True
            options["minimum_velocity"] = 0.0

        return options

    def solver_options(self):
        options = super().solver_options()
        options["solver"] = "gurobi"
        gurobi_options = options["gurobi"] = {}
        if self._stage == 1:
            gurobi_options["MIPgap"] = 0.005
        else:
            gurobi_options["MIPgap"] = 0.02
        gurobi_options["threads"] = 4
        gurobi_options["LPWarmStart"] = 2

        return options

    def bounds(self):
        bounds = super().bounds()

        if self._stage == 2:
            bounds.update(self.__boolean_bounds)

        return bounds


class EndScenarioSizingStagedHIGHS(EndScenarioSizingStaged):
    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        if self._stage == 1:
            highs_options["mip_rel_gap"] = 0.005
        else:
            highs_options["mip_rel_gap"] = 0.02

        options["gurobi"] = None

        return options


class EndScenarioSizingCBC(EndScenarioSizing):
    def post(self):
        super().post()

        self._write_updated_esdl()

    def solver_options(self):
        options = super().solver_options()
        # options["casadi_solver"] = self._qpsol
        options["solver"] = "cbc"
        options["gurobi"] = None

        if options["solver"] == "cbc":
            options["hot_start"] = self._hot_start
            cbc_options = options["cbc"] = {}
            cbc_options["seconds"] = 300000.0

        return options


def run_end_scenario_sizing(end_scenario_problem_class, staged_pipe_optimization=True):
    """
    This function is used to run end_scenario_sizing problem. There are a few variations of the
    same basic class. The main functionality this function adds is the staged approach, where
    we first solve without heat_losses, to then solve the same problem with heat losses but
    constraining the problem to only allow for the earlier found pipe classes and one size up.

    This staged approach is done to speed up the problem, as the problem without heat losses is
    much faster as it avoids inequality big_m constraints for the heat to discharge on pipes. The
    one size up possibility is to avoid infeasibilities in compensating for the heat losses.

    Parameters
    ----------
    end_scenario_problem_class : The end scenario problem class.
    staged_pipe_optimization : Boolean to toggle between the staged or non-staged approach

    Returns
    -------

    """
    import time

    boolean_bounds = {}

    start_time = time.time()
    if staged_pipe_optimization:
        solution = run_optimization_problem(end_scenario_problem_class, stage=1)
        results = solution.extract_results()
        parameters = solution.parameters(0)

        # We give bounds for stage 2 by allowing one DN sizes larger than what was found in the
        # stage 1 optimization.
        pc_map = solution._HeatMixin__pipe_topo_pipe_class_map
        for pipe_classes in pc_map.values():
            v_prev = 0.0
            first_pipe_class = True
            for var_name in pipe_classes.values():
                v = results[var_name][0]
                if first_pipe_class and abs(v) == 1.0:
                    boolean_bounds[var_name] = (abs(v), abs(v))
                elif abs(v) == 1.0:
                    boolean_bounds[var_name] = (0.0, abs(v))
                elif v_prev == 1.0:
                    boolean_bounds[var_name] = (0.0, 1.0)
                else:
                    boolean_bounds[var_name] = (abs(v), abs(v))
                v_prev = v
                first_pipe_class = False

        for asset in [
            *solution.heat_network_components.get("source", []),
            *solution.heat_network_components.get("buffer", []),
        ]:
            var_name = f"{asset}_aggregation_count"
            lb = results[var_name][0]
            ub = solution.bounds()[var_name][1]
            if round(lb) >= 1:
                boolean_bounds[var_name] = (lb, ub)

        t = solution.times()
        from rtctools.optimization.timeseries import Timeseries

        for p in solution.heat_network_components.get("pipe", []):
            if p in solution.hot_pipes and parameters[f"{p}.area"] > 0.0:
                lb = []
                ub = []
                for i in range(len(t)):
                    r = results[f"{p}__flow_direct_var"][i]
                    # bound to roughly represent 4km of heat losses in pipes
                    lb.append(
                        r if abs(results[f"{p}.Q"][i] / parameters[f"{p}.area"]) > 2.5e-2 else 0
                    )
                    ub.append(
                        r if abs(results[f"{p}.Q"][i] / parameters[f"{p}.area"]) > 2.5e-2 else 1
                    )

                boolean_bounds[f"{p}__flow_direct_var"] = (Timeseries(t, lb), Timeseries(t, ub))
                try:
                    r = results[f"{p}__is_disconnected"]
                    boolean_bounds[f"{p}__is_disconnected"] = (Timeseries(t, r), Timeseries(t, r))
                except KeyError:
                    pass

    solution = run_optimization_problem(
        end_scenario_problem_class,
        stage=2,
        boolean_bounds=boolean_bounds,
    )

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))

    return solution


@main_decorator
def main(runinfo_path, log_level):
    logger.info("Run Scenario Sizing")

    kwargs = {
        "write_result_db_profiles": False,
        "influxdb_host": "localhost",
        "influxdb_port": 8086,
        "influxdb_username": None,
        "influxdb_password": None,
        "influxdb_ssl": False,
        "influxdb_verify_ssl": False,
    }

    _ = run_optimization_problem(
        EndScenarioSizingHIGHS,
        esdl_run_info_path=runinfo_path,
        log_level=log_level,
        **kwargs,
    )
    # results = solution.extract_results()


if __name__ == "__main__":
    main()



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestAtes(TestCase):
    def test_ates(self):
        """
        This test checks the constraints concerning the heat2discharge and energy conservation for
        the ates. THe heat loss model used annd the typical cyclic constraint that will be applied
        in most use cases.
        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with 365 day timesteps
        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        stored_heat = results["ATES_033c.Stored_heat"]
        heat_loss = results["ATES_033c.Heat_loss"]
        heat_ates = results["ATES_033c.Heat_ates"]
        summed_charge = np.sum(np.clip(heat_ates, 0.0, np.inf))
        summed_discharge = np.abs(np.sum(np.clip(heat_ates, -np.inf, 0.0)))

        # Test if the heat loss is as expected
        coeff = solution.parameters(0)["ATES_033c.heat_loss_coeff"]
        np.testing.assert_allclose(heat_loss, coeff * stored_heat)
        np.testing.assert_array_less(summed_discharge, summed_charge)

        # Test begin and end same state
        np.testing.assert_allclose(stored_heat[0], stored_heat[-1])

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestAtes()
    a.test_ates()
    temp = 0



from pathlib import Path
from unittest import TestCase

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestAbsoluteHeat(TestCase):
    def test_absolute_heat(self):
        """
        This is a single line ring model, meaning that there are no dedicated supply or return
        lines. This means that this test checks whether this model is correctly parsed from esdl
        and meets the logical energy and heat to discharge constraints.
        """
        import models.absolute_heat.src.example as example
        from models.absolute_heat.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_optimization_problem(HeatProblem, base_folder=base_folder)

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class ControlValve(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "control_valve"

        self.add_variable(Variable, "dH")

        self.add_equation(self.dH - (self.QTHOut.H - self.QTHIn.H))

        self.add_equation(self.QTHIn.T - self.QTHOut.T)



from dataclasses import dataclass


@dataclass(frozen=True)
class DemandInsulationClass:
    """
    This class defines the properties for an insulation level/standard. The optimizer can
    select out of a list of insulation levels for each demand. There is a cost associated to
    changing a demand from its existing insulation level to another insulation level. Once the
    insulation is installed it will affect the demand profile and thus the amount of energy that is
    consumed. It will also affect the required temperature at the demand which can have an impact
    on the feasible network temperature selected. For these reasons this class consists out of the
    following variables:
    - name_insulation_level: Defines the name of the insulation level
    - name_demand: Demand asset to which the insulation level can apply
    - minimum_temperature_deg: Is the minimum temperature needed at the demand in unit deg.
    - demand_scaling_factor: A scalar that is used to re-scale the demand to represent the demand
    at that insulation level
    - investment_cost_euro: Is the amount of cost associated with reaching this insulation level.
    """

    name_insulation_level: str
    name_demand: str
    minimum_temperature_deg: float
    demand_scaling_factor: float  # value between 0.0 and 1.0
    investment_cost_euro: float



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.head_loss_mixin import HeadLossOption


class TestHeat(TestCase):
    def test_heat_loss(self):
        import models.double_pipe_heat.src.double_pipe_heat as double_pipe_heat
        from models.double_pipe_heat.src.double_pipe_heat import DoublePipeEqualHeat

        base_folder = Path(double_pipe_heat.__file__).resolve().parent.parent

        case = run_optimization_problem(DoublePipeEqualHeat, base_folder=base_folder)
        results = case.extract_results()

        source = results["source.Heat_source"]
        demand = results["demand.Heat_demand"]

        # With non-zero heat losses in pipes, the demand should always be
        # strictly lower than what is produced.
        np.testing.assert_array_less(demand, source)

    def test_zero_heat_loss(self):
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatPython

        class Model(HeatPython):
            def parameters(self, ensemble_member):
                parameters = super().parameters(ensemble_member)

                for pipe in self.heat_network_components["pipe"]:
                    assert f"{pipe}.Heat_loss" in parameters
                    parameters[f"{pipe}.Heat_loss"] = 0.0

                return parameters

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent

        run_optimization_problem(Model, base_folder=base_folder)


class TestMinMaxPressureOptions(TestCase):
    import models.basic_source_and_demand.src.heat_comparison as heat_comparison
    from models.basic_source_and_demand.src.heat_comparison import HeatPython

    base_folder = Path(heat_comparison.__file__).resolve().parent.parent
    min_pressure = 4.0
    max_pressure = 12.0

    class SmallerPipes(HeatPython):
        # We want to force the dynamic pressure in the system to be higher
        # than 12 - 4 = 8 bar (typical upper and lower bounds). We make the
        # pipes smaller in diameter/area to accomplish this. We also adjust
        # the minimum heat delivered by the source, such that a maximum
        # pressure range can force this to go lower than usual.
        def parameters(self, ensemble_member):
            parameters = super().parameters(ensemble_member)
            for p in self.heat_network_components["pipe"]:
                parameters[f"{p}.diameter"] = 0.04
                parameters[f"{p}.area"] = 0.25 * 3.14159265 * parameters[f"{p}.diameter"] ** 2
            return parameters

        def bounds(self):
            bounds = super().bounds()
            bounds["source.Heat_source"] = (0.0, 125_000.0)
            return bounds

    class MinPressure(SmallerPipes):
        def heat_network_options(self):
            options = super().heat_network_options()
            assert "pipe_minimum_pressure" in options
            options["pipe_minimum_pressure"] = TestMinMaxPressureOptions.min_pressure
            return options

    class MaxPressure(SmallerPipes):
        def heat_network_options(self):
            options = super().heat_network_options()
            assert "pipe_maximum_pressure" in options
            options["pipe_maximum_pressure"] = TestMinMaxPressureOptions.max_pressure
            return options

    class MinMaxPressure(SmallerPipes):
        def heat_network_options(self):
            options = super().heat_network_options()
            options["pipe_minimum_pressure"] = TestMinMaxPressureOptions.min_pressure
            options["pipe_maximum_pressure"] = TestMinMaxPressureOptions.max_pressure
            return options

    def test_min_max_pressure_options(self):
        case_default = run_optimization_problem(self.SmallerPipes, base_folder=self.base_folder)
        case_min_pressure = run_optimization_problem(self.MinPressure, base_folder=self.base_folder)
        case_max_pressure = run_optimization_problem(self.MaxPressure, base_folder=self.base_folder)
        case_min_max_pressure = run_optimization_problem(
            self.MinMaxPressure, base_folder=self.base_folder
        )

        def _get_min_max_pressure(case):
            min_head = np.inf
            max_head = -np.inf

            results = case.extract_results()
            for p in case.heat_network_components["pipe"]:
                min_head_in = min(results[f"{p}.HeatIn.H"])
                min_head_out = min(results[f"{p}.HeatOut.H"])
                min_head = min([min_head, min_head_in, min_head_out])

                max_head_in = max(results[f"{p}.HeatIn.H"])
                max_head_out = max(results[f"{p}.HeatOut.H"])
                max_head = max([max_head, max_head_in, max_head_out])

            return min_head / 10.2, max_head / 10.2

        min_, max_ = _get_min_max_pressure(case_default)
        self.assertGreater(max_ - min_, self.max_pressure - self.min_pressure)
        base_objective_value = case_default.objective_value

        min_, max_ = _get_min_max_pressure(case_min_pressure)
        self.assertGreater(min_, self.min_pressure * 0.99)
        self.assertGreater(max_, self.max_pressure)
        self.assertAlmostEqual(case_min_pressure.objective_value, base_objective_value, 4)

        min_, max_ = _get_min_max_pressure(case_max_pressure)
        self.assertLess(min_, self.min_pressure)
        self.assertLess(max_, self.max_pressure * 1.01)
        self.assertAlmostEqual(case_max_pressure.objective_value, base_objective_value, 4)

        min_, max_ = _get_min_max_pressure(case_min_max_pressure)
        self.assertGreater(min_, self.min_pressure * 0.99)
        self.assertLess(max_, self.max_pressure * 1.01)
        self.assertGreater(case_min_max_pressure.objective_value, base_objective_value * 1.5)


class TestDisconnectablePipe(TestCase):
    import models.basic_source_and_demand.src.heat_comparison as heat_comparison
    from models.basic_source_and_demand.src.heat_comparison import HeatPython

    base_folder = Path(heat_comparison.__file__).resolve().parent.parent

    class ModelConnected(HeatPython):
        # We allow the pipe to be disconnectable. We need to be sure that
        # the solution is still feasible (source delivering no heat), so we
        # lower the lower bound.

        def parameters(self, ensemble_member):
            parameters = super().parameters(ensemble_member)
            for p in self.heat_network_components["pipe"]:
                parameters[f"{p}.disconnectable"] = True
            return parameters

        def bounds(self):
            bounds = super().bounds()
            bounds["source.Heat_source"] = (0.0, 125_000.0)
            return bounds

    class ModelDisconnected(ModelConnected):
        def constraints(self, ensemble_member):
            constraints = super().constraints(ensemble_member)

            # Note that we still enforce a _minimum_ velocity in the pipe, if it
            # is connected. So if we force the discharge to zero, that means we
            # force it to be disconnected.
            times = self.times()
            q = self.state_at("pipe_hot.Q", times[1], ensemble_member)
            constraints.append((q, 0.0, 0.0))

            return constraints

        def heat_network_options(self):
            options = super().heat_network_options()
            options["heat_loss_disconnected_pipe"] = False
            return options

    class ModelDisconnectedNoHeatLoss(ModelDisconnected):
        def heat_network_options(self):
            options = super().heat_network_options()
            options["heat_loss_disconnected_pipe"] = False
            return options

    def test_disconnected_network_pipe(self):
        case_connected = run_optimization_problem(self.ModelConnected, base_folder=self.base_folder)
        results_connected = case_connected.extract_results()
        q_connected = results_connected["pipe_hot.Q"]

        case_disconnected = run_optimization_problem(
            self.ModelDisconnected, base_folder=self.base_folder
        )
        results_disconnected = case_disconnected.extract_results()
        q_disconnected = results_disconnected["pipe_hot.Q"]

        # Sanity check, as we rely on the minimum velocity being strictly
        # larger than zero for the discharge constraint to disconnect the
        # pipe.
        self.assertGreater(case_connected.heat_network_options()["minimum_velocity"], 0.0)

        self.assertLess(q_disconnected[1], q_connected[1])
        self.assertAlmostEqual(q_disconnected[1], 0.0, 5)

        np.testing.assert_allclose(q_connected[2:], q_disconnected[2:])

    class ModelDisconnectedDarcyWeisbach(ModelDisconnected):
        def heat_network_options(self):
            options = super().heat_network_options()
            options["head_loss_option"] = HeadLossOption.LINEARIZED_DW
            return options

    def test_disconnected_pipe_darcy_weisbach(self):
        """
        Just a sanity check that the head loss constraints for disconnectable
        pipes works with LINEAR as well as LINEARIZED_DW.
        """

        case_linear = run_optimization_problem(self.ModelDisconnected, base_folder=self.base_folder)
        results_linear = case_linear.extract_results()
        q_linear = results_linear["pipe_hot.Q"]

        case_dw = run_optimization_problem(
            self.ModelDisconnectedDarcyWeisbach, base_folder=self.base_folder
        )
        results_dw = case_dw.extract_results()
        q_dw = results_dw["pipe_hot.Q"]

        # Without any constraints on the maximum or minimum head/pressure
        # (loss) in the system, we expect equal results.
        np.testing.assert_allclose(q_linear, q_dw)



from pathlib import Path
from unittest import TestCase

import numpy as np


from rtctools.util import run_optimization_problem

# TODO: still have to make test where elecitricity direction is switched:
# e.g. 2 nodes, with at each node a producer and consumer, first one node medium demand, second
# small demand and then increase the demand of the second node such that direction changes


class TestMILPElectricSourceSink(TestCase):
    """Unit tests for the MILP test case of a source, a cable, a sink"""

    def test_source_sink(self):
        """Test to verify if when a target it set that is more or less than
        then max or min of the respective component it is capped. 1000"""
        import models.unit_cases_electricity.source_sink_cable.src.example as example
        from models.unit_cases_electricity.source_sink_cable.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent
        max_ = 1000.0  # This max is set in the esdl file
        v_min = 230  # set as minimum voltage for cables
        tol = 1e-10

        results = run_optimization_problem(
            ElectricityProblem, base_folder=base_folder
        ).extract_results()

        # Test if capping is ok
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        smallerthen = all(power_consumed <= np.ones(len(power_consumed)) * max_)
        self.assertTrue(smallerthen)
        biggerthen = all(power_consumed >= np.zeros(len(power_consumed)))
        self.assertTrue(biggerthen)

        # Test energy conservation
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        power_delivered = results["ElectricityProducer_b95d.ElectricityOut.Power"]
        power_loss = results["ElectricityCable_238f.Power_loss"]
        total_power_dissipation = power_consumed + power_loss
        self.assertIsNone(
            np.testing.assert_allclose(total_power_dissipation, power_delivered, rtol=1e-4),
            msg="No energy conservation. Total demand is not equal to total delivery.",
        )
        biggerthen = all(power_loss >= np.zeros(len(power_loss)))
        self.assertTrue(biggerthen)

        # Test that voltage goes down
        v_in = results["ElectricityCable_238f.ElectricityIn.V"]
        v_out = results["ElectricityCable_238f.ElectricityOut.V"]
        np.testing.assert_array_less(v_out, v_in)
        biggerthen = all(v_out >= (v_min - tol) * np.ones(len(v_out)))
        self.assertTrue(biggerthen)

    def test_source_sink_max_curr(self):
        """Test to verify if when a target it set that is more or less than
        then max or min of the respective component it is capped. 1000"""
        import models.unit_cases_electricity.source_sink_cable.src.example as example
        from models.unit_cases_electricity.source_sink_cable.src.example import (
            ElectricityProblemMaxCurr,
        )

        base_folder = Path(example.__file__).resolve().parent.parent
        max_ = 142.0 * 1.5e4  # This max is based on max current and voltage requirement at consumer
        v_min = 1.0e4  # set as minimum voltage for cables

        results = run_optimization_problem(
            ElectricityProblemMaxCurr, base_folder=base_folder
        ).extract_results()

        tolerance = 1e-10  # due to computational comparison

        # Test if capping is ok (capping based on max power as result of v_min*Imax)
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        smallerthen = all(power_consumed - tolerance <= np.ones(len(power_consumed)) * max_)
        self.assertTrue(smallerthen)
        biggerthen = all(power_consumed >= np.zeros(len(power_consumed)))
        self.assertTrue(biggerthen)

        # Test energy conservation
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        power_delivered = results["ElectricityProducer_b95d.ElectricityOut.Power"]
        power_loss = results["ElectricityCable_238f.Power_loss"]
        total_power_dissipation = power_consumed + power_loss
        self.assertIsNone(
            np.testing.assert_allclose(total_power_dissipation, power_delivered, rtol=1e-4),
            msg="No energy conservation. Total demand is not equal to total delivery.",
        )
        biggerthen = all(power_loss >= np.zeros(len(power_loss)))
        self.assertTrue(biggerthen)

        # Test that voltage goes down
        v_in = results["ElectricityCable_238f.ElectricityIn.V"]
        v_out = results["ElectricityCable_238f.ElectricityOut.V"]
        np.testing.assert_array_less(v_out, v_in)
        biggerthen = all(v_out >= v_min * np.ones(len(v_out)) - tolerance)
        self.assertTrue(biggerthen)

        # Test that max current is not exceeded and is constant along path (since no nodes included)
        current_demand = results["ElectricityDemand_2af6.ElectricityIn.I"]
        current_producer = results["ElectricityProducer_b95d.ElectricityOut.I"]
        current_cable = results["ElectricityCable_238f.ElectricityOut.I"]
        np.testing.assert_allclose(current_demand, current_cable)
        np.testing.assert_allclose(current_cable, current_producer)
        biggerthen = all(142.0 * np.ones(len(current_demand)) >= current_demand - tolerance)
        self.assertTrue(biggerthen)



from pathlib import Path
from unittest import TestCase

import numpy as np


from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestProducerMaxProfile(TestCase):
    """To verify that the producer can have a given scaled profile, where the producer will always
    produce equal or less than said profile.
    The constraint for the producer profile is checked by ensuring that the producer is temporarily
    less available (reducing the profile value at a few time steps)."""

    def test_max_producer_profile(self):
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemProdProfile

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblemProdProfile, base_folder=base_folder)
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)
        tol = 1e-8
        heat_demand1 = results["HeatingDemand_a3b8.Heat_demand"]
        heat_producer = results["GeothermalSource_b702.Heat_source"]
        size_producer = results["GeothermalSource_b702__max_size"]

        heatdemand1_target = solution.get_timeseries("HeatingDemand_a3b8.target_heat_demand").values
        heat_producer_profile_scaled = solution.get_timeseries(
            "GeothermalSource_b702.target_heat_source"
        ).values
        heat_producer_profile_full = heat_producer_profile_scaled * size_producer

        # check that heat produced is smaller than the profile
        biggerthen = all(heat_producer_profile_full + tol >= heat_producer)
        self.assertTrue(biggerthen)

        # heat should be stored in buffer and thus heatdemand should still be able to be matched.
        np.testing.assert_allclose(heat_demand1, heatdemand1_target)



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestHEX(TestCase):
    def test_heat_exchanger(self):
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        prim_heat = results["HeatExchange_39ed.Primary_heat"]
        sec_heat = results["HeatExchange_39ed.Secondary_heat"]
        disabled = results["HeatExchange_39ed__disabled"]

        # We check the energy converted betweeen the commodities
        eff = solution.parameters(0)["HeatExchange_39ed.efficiency"]

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        np.testing.assert_allclose(prim_heat * eff, sec_heat)

        # Note that we are not testing the last element as we exploit the last timestep for
        # checking the disabled boolean and the assert statement doesn't work for a difference of
        # zero
        np.testing.assert_allclose(prim_heat[-1], 0.0, atol=1e-8)
        np.testing.assert_allclose(disabled[-1], 1.0)
        np.testing.assert_allclose(disabled[:-1], 0.0)
        # Check that heat is flowing through the hex
        np.testing.assert_array_less(-prim_heat[:-1], 0.0)


class TestHP(TestCase):
    def test_heat_pump(self):
        import models.heatpump.src.run_heat_pump as run_heat_pump
        from models.heatpump.src.run_heat_pump import (
            HeatProblem,
        )

        base_folder = Path(run_heat_pump.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        prim_heat = results["GenericConversion_3d3f.Primary_heat"]
        sec_heat = results["GenericConversion_3d3f.Secondary_heat"]
        power_elec = results["GenericConversion_3d3f.Power_elec"]

        # TODO: we should also check if heatdemand target is matched
        # TODO: check if the primary source utilisisation is maximised and secondary minimised
        # I think it is badly scaled or some scaling issues in constraints. (possible in combination
        # with disconnected)

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        # We check the energy converted betweeen the commodities
        np.testing.assert_allclose(power_elec * 4.0, sec_heat)
        np.testing.assert_allclose(power_elec + prim_heat, sec_heat)



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestMaxSizeAggregationCount(TestCase):
    def test_max_size_and_aggr_count(self):
        import models.test_case_small_network_with_ates_with_buffer.src.run_ates as run_ates
        from models.test_case_small_network_with_ates_with_buffer.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a few days
        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        # Producer 1 should not produce due to higher cost
        # Producer 2 should produce
        heat_1 = results["HeatProducer_1.Heat_source"]
        heat_2 = results["HeatProducer_2.Heat_source"]
        prod_1_placed = results["HeatProducer_1_aggregation_count"]
        prod_2_placed = results["HeatProducer_2_aggregation_count"]
        var_cost_1 = results["HeatProducer_1__variable_operational_cost"]
        var_cost_2 = results["HeatProducer_2__variable_operational_cost"]
        inst_cost_1 = results["HeatProducer_1__installation_cost"]
        inst_cost_2 = results["HeatProducer_2__installation_cost"]
        inv_cost_1 = results["HeatProducer_1__investment_cost"]
        inv_cost_2 = results["HeatProducer_2__investment_cost"]
        max_size_1 = results["HeatProducer_1__max_size"]
        max_size_2 = results["HeatProducer_2__max_size"]

        # Test if source 1 is off and 2 is producing
        np.testing.assert_allclose(heat_1, 0.0, atol=1e-6)
        np.testing.assert_equal(True, heat_2[1:] > 0.0)

        # Test if source 1 is not placed and 2 is placed
        np.testing.assert_allclose(prod_1_placed, 0.0)
        np.testing.assert_allclose(prod_2_placed, 1.0, atol=1.0e-6)

        # Test that max size is correct, note that we use an equality check as due to the cost
        # minimization they should be equal.
        np.testing.assert_allclose(max(heat_2), max_size_2)
        np.testing.assert_allclose(max_size_1, 0.0)

        # Test that investmentcost is correctly linked to max size
        np.testing.assert_allclose(
            inv_cost_2,
            solution.parameters(0)["HeatProducer_2.investment_cost_coefficient"] * max_size_2,
        )

        # Test that cost only exist for 2 and not for 1. Note the tolerances
        # to avoid test failing when heat losses slightly change
        np.testing.assert_allclose(var_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(var_cost_2, 6174.920222, atol=1000.0, rtol=1.0e-2)
        np.testing.assert_allclose(inst_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(inst_cost_2, 100000.0)
        np.testing.assert_allclose(inv_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(inv_cost_2, 476459.893686, atol=1000.0, rtol=1.0e-2)

        # Since the buffer and ates are not optional they must consume some heat to compensate
        # losses as the buffer has a minimum fraction volume of 5%.
        # Therefore, we can check the max_size constraint.
        np.testing.assert_allclose(
            True, results["HeatStorage_74c1.Stored_heat"] <= results["HeatStorage_74c1__max_size"]
        )
        np.testing.assert_allclose(
            True, abs(results["ATES_033c.Heat_ates"]) <= results["ATES_033c__max_size"]
        )
        np.testing.assert_allclose(results["ATES_033c_aggregation_count"], 1.0)
        np.testing.assert_allclose(results["HeatStorage_74c1_aggregation_count"], 1.0)

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates
        from models.test_case_small_network_ates_buffer_optional_assets.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is the same problem, but now with the buffer and ates also optional.
        # Therefore, we expect that the ates and buffer are no longer placed to avoid their heat
        # losses. This allows us to check if their placement constraints are proper.
        solution = run_optimization_problem(
            HeatProblem,
            base_folder=base_folder,
        )

        results = solution.extract_results()

        np.testing.assert_allclose(results["ATES_033c.Heat_ates"], 0.0, atol=1.0e-6)
        np.testing.assert_allclose(results["HeatStorage_74c1.Stored_heat"], 0.0, atol=1.0e-3)
        np.testing.assert_allclose(results["ATES_033c_aggregation_count"], 0.0, atol=1.0e-6)
        np.testing.assert_allclose(results["HeatStorage_74c1_aggregation_count"], 0.0, atol=1.0e-6)

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)



import logging
import numbers
import os
import sys
import traceback
import uuid
from pathlib import Path

import esdl
from esdl.profiles.influxdbprofilemanager import ConnectionSettings
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager
from esdl.profiles.profilemanager import ProfileManager

# from esdl.profiles.excelprofilemanager import ExcelProfileManager

import numpy as np

import pandas as pd

import pytz

from rtctools_heat_network.constants import GRAVITATIONAL_CONSTANT
from rtctools_heat_network.esdl.edr_pipe_class import EDRPipeClass
from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.workflows.utils.helpers import _sort_numbered


logger = logging.getLogger("rtctools_heat_network")


class ScenarioOutput(HeatMixin):
    __optimized_energy_system_handler = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Settings for influxdb when writing out result profile data to it
        # Default settings
        self.write_result_db_profiles = False
        self.influxdb_username = None
        self.influxdb_password = None

        base_error_string = "Missing influxdb setting for writing result profile data:"
        try:
            self.write_result_db_profiles = kwargs["write_result_db_profiles"]

            if self.write_result_db_profiles not in [True, False]:
                logger.error(
                    "Current setting of write_result_db_profiles is: "
                    f"{self.write_result_db_profiles} and it should be set to True or False"
                )
                sys.exit(1)

            if self.write_result_db_profiles:
                try:
                    self.influxdb_host = kwargs["influxdb_host"]
                    if len(self.influxdb_host) == 0:
                        logger.error(
                            "Current setting of influxdb_host is an empty string and it should"
                            " be the name of the host"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} host")
                    sys.exit(1)
                try:
                    self.influxdb_port = kwargs["influxdb_port"]
                    if not isinstance(self.influxdb_port, int):
                        logger.error(
                            "Current setting of influxdb_port is: "
                            f"{self.influxdb_port} and it should be set to int value (port number)"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} port")
                    sys.exit(1)
                try:
                    self.influxdb_username = kwargs["influxdb_username"]
                except KeyError:
                    logger.error(f"{base_error_string} username")
                    sys.exit(1)
                try:
                    self.influxdb_password = kwargs["influxdb_password"]
                except KeyError:
                    logger.error(f"{base_error_string} password")
                    sys.exit(1)
                try:
                    self.influxdb_ssl = kwargs["influxdb_ssl"]
                    if self.influxdb_ssl not in [True, False]:
                        logger.error(
                            "Current setting of influxdb_ssl is: "
                            f"{self.influxdb_ssl} and it should be set to True or False"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} ssl")
                    sys.exit(1)
                try:
                    self.influxdb_verify_ssl = kwargs["influxdb_verify_ssl"]
                    if self.influxdb_verify_ssl not in [True, False]:
                        logger.error(
                            "Current setting of influxdb_verify_ssl is: "
                            f"{self.influxdb_verify_ssl} and it should be set to True or False"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error("f{base_string} verify_ssl")
                    sys.exit(1)
        except KeyError:
            # Not writing out to a influxdb, so no settings are requried
            pass

    def get_optimized_esh(self):
        return self.__optimized_energy_system_handler

    def _write_html_output(self, template_name="mpc_buffer_sizing_output"):
        from jinja2 import Environment, FileSystemLoader

        assert self.ensemble_size == 1

        results = self.extract_results()
        parameters = self.parameters(0)

        # Format the priority results
        priority_results = [
            dict(
                number=number,
                success=success,
                pretty_time=f"{int(seconds // 60):02d}:{seconds % 60:06.3f}",
                objective_value=objective_value,
                return_status=stats["return_status"],
                secondary_return_status=stats.get("secondary_return_status", ""),
            )
            for (
                number,
                seconds,
                success,
                objective_value,
                stats,
            ) in self._priorities_output
        ]

        # Format the buffer results
        results_buffers_placed = {}
        results_buffers_size = {}
        results_sources_placed = {}
        results_sources_size = {}
        results_max_charging_rate = {}
        results_max_discharging_rate = {}

        for buffer in _sort_numbered(self.heat_network_components.get("buffer", [])):
            if buffer in self._minimize_size_buffers:
                max_size_var = self._max_buffer_heat_map[buffer]
                results_buffers_size[buffer] = float(results[max_size_var][0]) / (
                    parameters[f"{buffer}.cp"]
                    * parameters[f"{buffer}.rho"]
                    * (parameters[f"{buffer}.T_supply"] - parameters[f"{buffer}.T_return"])
                )
            else:
                results_buffers_size[buffer] = "-"

            if buffer in self._optional_buffers:
                buffer_placement_var = self._buffer_placement_map[buffer]
                results_buffers_placed[buffer] = np.round(results[buffer_placement_var][0]) == 1.0
            else:
                results_buffers_placed[buffer] = "-"

            (_, hot_orient), _ = self.heat_network_topology.buffers[buffer]
            q = hot_orient * results[f"{buffer}.HeatIn.Q"]
            inds_charging = q > 0
            inds_discharging = q < 0

            results_max_charging_rate[buffer] = max(q[inds_charging]) if any(inds_charging) else 0.0
            results_max_discharging_rate[buffer] = (
                max(-1 * q[inds_discharging]) if any(inds_discharging) else 0.0
            )

        buffer_results = [
            dict(
                name=buffer,
                tune_size=buffer in self._minimize_size_buffers,
                tune_placement=buffer in self._optional_buffers,
                maximum_size=self._override_buffer_size[buffer],
                result_placed=results_buffers_placed[buffer],
                result_size=results_buffers_size[buffer],
                max_charging_rate=results_max_charging_rate[buffer],
                max_discharging_rate=results_max_discharging_rate[buffer],
            )
            for buffer in self.heat_network_components.get("buffer", [])
        ]

        for source in _sort_numbered(self.heat_network_components["source"]):
            if source in self._minimize_size_sources:
                max_size_var = self._max_source_heat_map[source]
                results_sources_size[source] = float(results[max_size_var][0]) / 10.0**3
            else:
                results_sources_size[source] = "-"

            if source in self._optional_sources:
                source_placement_var = self._source_placement_map[source]
                results_sources_placed[source] = np.round(results[source_placement_var][0]) == 1.0
            else:
                results_sources_placed[source] = "-"

        source_results = [
            dict(
                name=source,
                tune_size=source in self._minimize_size_sources,
                tune_placement=source in self._optional_sources,
                maximum_size=self._override_max_power[source],
                result_placed=results_sources_placed[source],
                result_size=results_sources_size[source],
            )
            for source in self.heat_network_components["source"]
        ]

        # Format the pipe results
        # Note that we do not distinguish between routing and sizing
        # internally, but for the sake of the output we do.
        pipe_results = []

        for p in _sort_numbered(self.hot_pipes):
            pipe_classes = self.pipe_classes(p)
            tune_routing = len([pc for pc in pipe_classes if pc.inner_diameter == 0.0]) == 1
            inner_diameter = parameters[f"{p}.diameter"]
            asset = next(a for a in self.esdl_assets.values() if a.name == p)
            esdl_diameter = asset.attributes["innerDiameter"]

            if len(pipe_classes) <= 1:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "-"
                result_dn_size = "-"
            elif len(pipe_classes) == 2 and tune_routing:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_dn_size = "-"
            else:
                sorted_pipe_classes = sorted(
                    [pc for pc in pipe_classes if pc.inner_diameter > 0],
                    key=lambda pc: pc.inner_diameter,
                )

                tune_size = True
                min_dn_size = sorted_pipe_classes[0].name
                max_dn_size = sorted_pipe_classes[-1].name
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_pipe_class = self.get_optimized_pipe_class(p)
                result_dn_size = (
                    result_pipe_class.name if result_pipe_class is not None else inner_diameter
                )

            pipe_results.append(
                dict(
                    name=p,
                    tune_size=tune_size,
                    tune_routing=tune_routing,
                    esdl_diameter=esdl_diameter,
                    min_dn_size=min_dn_size,
                    max_dn_size=max_dn_size,
                    result_placed=result_placed,
                    result_dn_size=result_dn_size,
                )
            )

        input_csv_tables = {
            os.path.basename(x): pd.read_csv(x).to_dict("records")
            for x in self._csv_input_parameter_files
        }

        # Actually write out the html file based on the template
        templates_dir = Path(__file__).parent / "templates"

        env = Environment(loader=FileSystemLoader(templates_dir))
        template = env.get_template(template_name + ".html")

        os.makedirs(self._html_output_dir, exist_ok=True)

        filename = self._html_output_dir / (template_name + ".html")

        with open(filename, "w", encoding="utf-8") as fh:
            fh.write(
                template.render(
                    buffer_results=buffer_results,
                    source_results=source_results,
                    pipe_results=pipe_results,
                    priority_results=priority_results,
                    input_csv_tables=input_csv_tables,
                )
            )

    def _write_updated_esdl(self, optimizer_sim=True):
        from esdl.esdl_handler import EnergySystemHandler
        from rtctools_heat_network.esdl.esdl_mixin import _RunInfoReader

        results = self.extract_results()
        parameters = self.parameters(0)

        esh = EnergySystemHandler()
        if self.esdl_string is None:
            run_info = _RunInfoReader(self.esdl_run_info_path)
            energy_system = esh.load_file(str(run_info.esdl_file))
        else:
            energy_system = esh.load_from_string(self.esdl_string)

        input_energy_system_id = energy_system.id
        energy_system.id = str(uuid.uuid4())
        if optimizer_sim:
            energy_system.name = energy_system.name + "_GrowOptimized"
        else:
            energy_system.name = energy_system.name + "_Simulation"

        def _name_to_asset(name):
            return next(
                (x for x in energy_system.eAllContents() if hasattr(x, "name") and x.name == name)
            )

        # ------------------------------------------------------------------------------------------
        # KPIs
        # General cost breakdowns
        # ------------------------------------------------------------------------------------------
        kpis_top_level = esdl.KPIs(id=str(uuid.uuid4()))
        heat_source_energy_wh = {}
        asset_capex_breakdown = {}
        asset_opex_breakdown = {}
        tot_install_cost_euro = 0.0
        tot_invest_cost_euro = 0.0
        tot_variable_opex_cost_euro = 0.0
        tot_fixed_opex_cost_euro = 0.0

        for _key, asset in self.esdl_assets.items():
            asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
            placed = np.round(results[asset_placement_var][0]) >= 1.0
            if placed:
                try:
                    asset_capex_breakdown[asset.asset_type] += (
                        results[f"{asset.name}__installation_cost"][0]
                        + results[f"{asset.name}__investment_cost"][0]
                    )
                    tot_install_cost_euro += results[f"{asset.name}__installation_cost"][0]
                    tot_invest_cost_euro += results[f"{asset.name}__investment_cost"][0]

                    if (
                        results[f"{asset.name}__variable_operational_cost"][0] > 0.0
                        or results[f"{asset.name}__fixed_operational_cost"][0] > 0.0
                    ):
                        asset_opex_breakdown[asset.asset_type] += (
                            results[f"{asset.name}__variable_operational_cost"][0]
                            + results[f"{asset.name}__fixed_operational_cost"][0]
                        )

                        tot_variable_opex_cost_euro += results[
                            f"{asset.name}__variable_operational_cost"
                        ][0]
                        tot_fixed_opex_cost_euro += results[
                            f"{asset.name}__fixed_operational_cost"
                        ][0]

                except KeyError:
                    try:
                        asset_capex_breakdown[asset.asset_type] = (
                            results[f"{asset.name}__installation_cost"][0]
                            + results[f"{asset.name}__investment_cost"][0]
                        )
                        tot_install_cost_euro += results[f"{asset.name}__installation_cost"][0]
                        tot_invest_cost_euro += results[f"{asset.name}__investment_cost"][0]

                        if (
                            results[f"{asset.name}__variable_operational_cost"][0] > 0.0
                            or results[f"{asset.name}__fixed_operational_cost"][0] > 0.0
                        ):
                            asset_opex_breakdown[asset.asset_type] = (
                                results[f"{asset.name}__variable_operational_cost"][0]
                                + results[f"{asset.name}__fixed_operational_cost"][0]
                            )

                            tot_variable_opex_cost_euro += results[
                                f"{asset.name}__variable_operational_cost"
                            ][0]
                            tot_fixed_opex_cost_euro += results[
                                f"{asset.name}__fixed_operational_cost"
                            ][0]
                    except KeyError:
                        # Do not add any costs. Items like joint
                        pass

                # TODO: show discharge energy (current display) and charge energy
                # (new display to be added for ATES etc)
                if (
                    asset.asset_type == "HeatProducer"
                    or asset.asset_type == "GenericProducer"
                    or asset.asset_type == "ResidualHeatSource"
                    or asset.asset_type == "GeothermalSource"
                    or asset.asset_type == "ResidualHeatSource"
                    or asset.asset_type == "GasHeater"
                ):
                    heat_source_energy_wh[asset.name] = np.sum(
                        results[f"{asset.name}.Heat_source"][1:]
                        * (self.times()[1:] - self.times()[0:-1])
                        / 3600
                    )
                # TODO: ATES, HEAT pump show Secondary_heat and Primary_heat and tank storage
                # elif ATES:
                #  summed_charge = np.sum(np.clip(heat_ates, 0.0, np.inf))
                #  summed_discharge = np.abs(np.sum(np.clip(heat_ates, -np.inf, 0.0)))
                # elif Heat pump
                # elif asset.asset_type == "HeatStorage":  # Heat discharged
                #     heat_source_energy_wh[asset.name] = np.sum(
                #         np.clip(results[f"{asset.name}.Heat_buffer"][1:], -np.inf, 0.0)
                #         * (self.times()[1:] - self.times()[0:-1])
                #         / 3600
                #     )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="High level cost breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(
                            label="CAPEX",
                            value=tot_install_cost_euro + tot_invest_cost_euro,
                        ),
                        esdl.StringItem(
                            label="OPEX",
                            value=tot_variable_opex_cost_euro + tot_fixed_opex_cost_euro,
                        ),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Overall cost breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label="Installation", value=tot_install_cost_euro),
                        esdl.StringItem(label="Investment", value=tot_invest_cost_euro),
                        esdl.StringItem(label="Variable OPEX", value=tot_variable_opex_cost_euro),
                        esdl.StringItem(label="Fixed OPEX", value=tot_fixed_opex_cost_euro),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="CAPEX breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in asset_capex_breakdown.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="OPEX breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in asset_opex_breakdown.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Energy production [Wh]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in heat_source_energy_wh.items()
                    ]
                ),
            )
        )
        energy_system.instance[0].area.KPIs = kpis_top_level
        # ------------------------------------------------------------------------------------------
        # Cost breakdowns per polygon areas (can consist of several assets of differents types)
        # Notes:
        # - OPEX KPIs are taken into account for energy sources only.
        # - We assume that all energy produced outside of the the subarea comes in via a heat
        #   exchanger that is part of the subarea.
        # TODO: Investigate if no cost in the ESDL then this breaks ESDL visibility
        total_energy_produced_locally_wh = {}
        total_energy_consumed_locally_wh = {}
        estimated_energy_from_local_source_perc = {}
        estimated_energy_from_regional_source_perc = {}

        for subarea in energy_system.instance[0].area.area:
            area_investment_cost = 0.0
            area_installation_cost = 0.0
            area_variable_opex_cost = 0.0
            area_fixed_opex_cost = 0.0

            kpis = esdl.KPIs(id=str(uuid.uuid4()))
            # Here we make a breakdown of the produced energy in the subarea. Where we assume that
            # all energy produced outside of the the subarea comes in via a heat exchanger that is
            # part of the subarea.
            energy_breakdown = {}
            for asset in subarea.asset:
                asset_name = asset.name
                asset_type = self.get_asset_from_asset_name(asset_name).asset_type

                asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0

                if placed:
                    if asset_type == "Joint":
                        continue
                    try:
                        energy_breakdown[asset_type] += np.sum(results[f"{asset_name}.Heat_source"])
                    except KeyError:
                        try:
                            energy_breakdown[asset_type] = np.sum(
                                results[f"{asset_name}.Heat_source"]
                            )
                        except KeyError:
                            try:
                                energy_breakdown[asset_type] += np.sum(
                                    results[f"{asset_name}.Secondary_heat"]
                                )
                            except KeyError:
                                try:
                                    energy_breakdown[asset_type] = np.sum(
                                        results[f"{asset_name}.Secondary_heat"]
                                    )
                                except KeyError:
                                    pass

                    # Create KPIs by using applicable costs for the specific asset
                    area_investment_cost += results[self._asset_investment_cost_map[asset_name]][0]
                    area_installation_cost += results[
                        self._asset_installation_cost_map[asset_name]
                    ][0]
                    area_variable_opex_cost += results[
                        self._asset_variable_operational_cost_map[asset_name]
                    ][0]
                    area_fixed_opex_cost += results[
                        self._asset_fixed_operational_cost_map[asset_name]
                    ][0]

                    # Calculate the total energy [Wh] consumed/produced in an are.
                    # Note: heat losses of buffers, ATES' and pipes are included in the area energy
                    # consumption
                    if asset_name in self.heat_network_components.get("source", []):
                        try:
                            total_energy_produced_locally_wh[subarea.name] += np.sum(
                                results[f"{asset_name}.Heat_source"][1:]
                                * (self.times()[1:] - self.times()[0:-1])
                                / 3600.0
                            )
                        except KeyError:
                            total_energy_produced_locally_wh[subarea.name] = np.sum(
                                results[f"{asset_name}.Heat_source"][1:]
                                * (self.times()[1:] - self.times()[0:-1])
                                / 3600.0
                            )
                    if asset_name in self.heat_network_components.get("demand", []):
                        flow_variable = results[f"{asset_name}.Heat_demand"][1:]
                    elif asset_name in self.heat_network_components.get("buffer", []):
                        flow_variable = results[f"{asset_name}.Heat_buffer"][1:]
                    elif asset_name in self.heat_network_components.get("ates", []):
                        flow_variable = results[f"{asset_name}.Heat_ates"][1:]
                    elif asset_name in self.heat_network_components.get("pipe", []):
                        flow_variable = (
                            np.ones(len(self.times())) * results[f"{asset_name}__hn_heat_loss"]
                        )
                    else:
                        flow_variable = ""
                    if (
                        asset_name in self.heat_network_components.get("demand", [])
                        or asset_name in self.heat_network_components.get("buffer", [])
                        or asset_name in self.heat_network_components.get("ates", [])
                        or asset_name in self.heat_network_components.get("pipe", [])
                    ):
                        try:
                            total_energy_consumed_locally_wh[subarea.name] += np.sum(
                                flow_variable * (self.times()[1:] - self.times()[0:-1]) / 3600.0
                            )
                        except KeyError:
                            total_energy_consumed_locally_wh[subarea.name] = np.sum(
                                flow_variable * (self.times()[1:] - self.times()[0:-1]) / 3600.0
                            )
                    # end Calculate the total energy consumed/produced in an area
                # end if placed loop
            # end asset loop

            # Calculate the estimated energy source [%] for an area
            try:
                total_energy_produced_locally_wh_area = total_energy_produced_locally_wh[
                    subarea.name
                ]
            except KeyError:
                total_energy_produced_locally_wh_area = 0.0

            try:
                estimated_energy_from_local_source_perc[subarea.name] = min(
                    total_energy_produced_locally_wh_area
                    / total_energy_consumed_locally_wh[subarea.name]
                    * 100.0,
                    100.0,
                )
                estimated_energy_from_regional_source_perc[subarea.name] = min(
                    (100.0 - estimated_energy_from_local_source_perc[subarea.name]), 100.0
                )
            except KeyError:
                # Nothing to do, go on to next section of code
                pass

            # Here we add KPIs to the polygon area which allows to visualize them by hovering over
            # it with the mouse
            # Only update kpis if one of the costs > 0, else esdl file will be corrupted
            # TODO: discuss strange behaviour with Edwin - temporarily use of "True" in line below
            if area_investment_cost > 0.0 or area_installation_cost > 0.0 or True:
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_investment_cost / 1.0e6,
                        name="Investment",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_installation_cost / 1.0e6,
                        name="Installation",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
            # Only update kpis if one of the costs > 0, else esdl file will be corrupted
            # TODO: discuss strange behaviour with Edwin - temporarily use of "True" in line below
            if area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0 or True:
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_variable_opex_cost / 1.0e6,
                        name="Variable OPEX",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_fixed_opex_cost / 1.0e6,
                        name="Fixed OPEX",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )

            try:
                if total_energy_consumed_locally_wh[subarea.name] >= 0.0:
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(estimated_energy_from_local_source_perc[subarea.name], 1),
                            name="Estimated energy from local source(s) [%]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                unit=esdl.UnitEnum.PERCENT,
                                multiplier=esdl.MultiplierEnum.NONE,
                            ),
                        )
                    )
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(
                                estimated_energy_from_regional_source_perc[subarea.name], 1
                            ),
                            name="Estimated energy from regional source(s) [%]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                unit=esdl.UnitEnum.PERCENT,
                                multiplier=esdl.MultiplierEnum.NONE,
                            ),
                        )
                    )
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(total_energy_consumed_locally_wh[subarea.name] / 1.0e9, 1),
                            name="Total energy consumed [GWh]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                unit=esdl.UnitEnum.WATTHOUR,
                                multiplier=esdl.MultiplierEnum.GIGA,
                            ),
                        )
                    )
            except KeyError:
                # Do nothing because this area does not have any energy consumption
                pass

            # Create plots in the dashboard
            # Top level KPIs: Cost breakdown in a polygon area (for all assest grouped together)
            kpi_name = f"{subarea.name}: Asset cost breakdown [EUR]"
            if (area_installation_cost > 0.0 or area_investment_cost > 0.0) and (
                area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0
            ):
                polygon_area_string_item = [
                    esdl.StringItem(label="Installation", value=area_installation_cost),
                    esdl.StringItem(label="Investment", value=area_investment_cost),
                    esdl.StringItem(label="Variable OPEX", value=area_variable_opex_cost),
                    esdl.StringItem(label="Fixed OPEX", value=area_fixed_opex_cost),
                ]
            elif area_installation_cost > 0.0 or area_investment_cost > 0.0:
                polygon_area_string_item = [
                    esdl.StringItem(label="Installation", value=area_installation_cost),
                    esdl.StringItem(label="Investment", value=area_investment_cost),
                ]
            elif area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0:
                polygon_area_string_item = [
                    esdl.StringItem(label="Variable OPEX", value=area_variable_opex_cost),
                    esdl.StringItem(label="Fixed OPEX", value=area_fixed_opex_cost),
                ]
            if (
                area_installation_cost > 0.0
                or area_investment_cost > 0.0
                or area_variable_opex_cost > 0.0
                or area_fixed_opex_cost > 0.0
            ):
                kpis_top_level.kpi.append(
                    esdl.DistributionKPI(
                        name=kpi_name,
                        distribution=esdl.StringLabelDistribution(
                            stringItem=polygon_area_string_item
                        ),
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                        ),
                    )
                )

            # Here we add a distribution KPI to the subarea to which gives a piechart
            # !!!!!!!!!!!!!!! This will only work if the source is in the area?
            # TODO: Still to be resolved since piecharts are still a work in progress in mapeditor
            # kpis.kpi.append(
            #     esdl.DistributionKPI(
            #         name="Energy breakdown ?",
            #         distribution=esdl.StringLabelDistribution(
            #             stringItem=[
            #                 esdl.StringItem(label=key, value=value) for key,
            #                 value in energy_breakdown.items()
            #             ]
            #         )
            #     )
            # )
            subarea.KPIs = kpis
        # ebd sub-area loop

        # end KPIs
        # ------------------------------------------------------------------------------------------
        # Placement
        for _, attributes in self.esdl_assets.items():
            name = attributes.name
            if name in [
                *self.heat_network_components.get("source", []),
                *self.heat_network_components.get("ates", []),
                *self.heat_network_components.get("buffer", []),
            ]:
                asset = _name_to_asset(name)
                asset_placement_var = self._asset_aggregation_count_var_map[name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0
                max_size = results[self._asset_max_size_map[name]][0]

                if asset in self.heat_network_components.get("buffer", []):
                    asset.capacity = max_size
                    asset.volume = max_size / (
                        parameters[f"{name}.cp"]
                        * parameters[f"{name}.rho"]
                        * parameters[f"{name}.dT"]
                    )
                else:
                    asset.power = max_size
                if not placed:
                    asset.delete(recursive=True)
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED

        # Pipes:
        edr_pipe_properties_to_copy = ["innerDiameter", "outerDiameter", "diameter", "material"]

        esh_edr = EnergySystemHandler()

        for pipe in self.hot_pipes:
            pipe_classes = self.pipe_classes(pipe)
            # When a pipe has not been optimized, enforce pipe to be shown in the simulator
            # ESDL.
            if not pipe_classes:
                if optimizer_sim:
                    continue
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED

            if optimizer_sim:
                pipe_class = self.get_optimized_pipe_class(pipe)
            cold_pipe = self.hot_to_cold_pipe(pipe)

            if parameters[f"{pipe}.diameter"] != 0.0 or any(np.abs(results[f"{pipe}.Q"]) > 1.0e-9):
                # if not isinstance(pipe_class, EDRPipeClass):
                #     assert pipe_class.name == f"{pipe}_orig"
                #     continue
                # print(results[f"{pipe}.Q"])
                # print(pipe + " has pipeclass: " + pipe_class.name )
                # print(pipe + f" has diameter: " + pipe_class.name)

                if optimizer_sim:
                    assert isinstance(pipe_class, EDRPipeClass)

                    asset_edr = esh_edr.load_from_string(pipe_class.xml_string)

                for p in [pipe, cold_pipe]:
                    asset = _name_to_asset(p)
                    asset.state = esdl.AssetStateEnum.ENABLED

                    try:
                        asset.costInformation.investmentCosts.value = pipe_class.investment_costs
                    except AttributeError:
                        pass
                        # do nothing, in the case that no costs have been specified for the return
                        # pipe in the mapeditor
                    except UnboundLocalError:
                        pass

                    if optimizer_sim:
                        for prop in edr_pipe_properties_to_copy:
                            setattr(asset, prop, getattr(asset_edr, prop))
            else:
                for p in [pipe, cold_pipe]:
                    asset = _name_to_asset(p)
                    asset.delete(recursive=True)

        # ------------------------------------------------------------------------------------------
        # Important: This code below must be placed after the "Placement" code. Reason: it relies
        # on unplaced assets being deleted.
        # ------------------------------------------------------------------------------------------
        # Write asset result profile data to database. The datbase is setup as follows:
        #   - The each time step is represented by a row of data, with columns datetime, field
        #     values
        #   - Database name: input esdl id
        #   - Measurment: asset name
        #   - Fields: profile value for the specific variable
        #   - Tags used as filters: output esdl id

        if self.write_result_db_profiles:
            logger.info("Writing asset result profile data to influxDB")
            results = self.extract_results()
            # Note: when adding new variables to variables_one_hydraulic_system or"
            # variables_two_hydraulic_system also add quantity and units to the ESDL for the new
            # variables in the code lower down
            variables_one_hydraulic_system = ["HeatIn.Q", "HeatIn.H", "Heat_flow"]
            variables_two_hydraulic_system = [
                "Primary.HeatIn.Q",
                "Primary.HeatIn.H",
                "Secondary.HeatIn.Q",
                "Secondary.HeatIn.H",
                "Heat_flow",
            ]

            influxdb_conn_settings = ConnectionSettings(
                host=self.influxdb_host,
                port=self.influxdb_port,
                username=self.influxdb_username,
                password=self.influxdb_password,
                database=input_energy_system_id,
                ssl=self.influxdb_ssl,
                verify_ssl=self.influxdb_verify_ssl,
            )

            for asset_name in [
                *self.heat_network_components.get("source", []),
                *self.heat_network_components.get("demand", []),
                *self.heat_network_components.get("pipe", []),
                *self.heat_network_components.get("buffer", []),
                *self.heat_network_components.get("ates", []),
                *self.heat_network_components.get("heat_exchanger", []),
                *self.heat_network_components.get("heat_pump", []),
            ]:
                profiles = ProfileManager()
                profiles.profile_type = "DATETIME_LIST"
                profiles.profile_header = ["datetime"]
                try:
                    # If the asset has been placed
                    asset = _name_to_asset(asset_name)

                    # Get index of outport which will be used to assign the profile data to
                    index_outport = -1
                    for ip in range(len(asset.port)):
                        if isinstance(asset.port[ip], esdl.OutPort):
                            if index_outport == -1:
                                index_outport = ip
                            else:
                                logger.warning(
                                    f"Asset {asset_name} has more than 1 OutPort, and the "
                                    "profile data has been assigned to the 1st OutPort"
                                )
                                break

                    if index_outport == -1:
                        logger.error(
                            f"Variable {index_outport} has not been assigned to the asset OutPort"
                        )
                        sys.exit(1)

                    for ii in range(len(self.times())):
                        if not self.io.datetimes[ii].tzinfo:
                            data_row = [pytz.utc.localize(self.io.datetimes[ii])]
                        else:
                            data_row = [self.io.datetimes[ii]]

                        try:
                            # For all components dealing with one hydraulic system
                            if isinstance(
                                results[f"{asset_name}." + variables_one_hydraulic_system[0]][ii],
                                numbers.Number,
                            ):
                                variables_names = variables_one_hydraulic_system
                        except Exception:
                            # For all components dealing with two hydraulic system
                            if isinstance(
                                results[f"{asset_name}." + variables_two_hydraulic_system[0]][ii],
                                numbers.Number,
                            ):
                                variables_names = variables_two_hydraulic_system

                        for variable in variables_names:
                            if ii == 0:
                                # Set header for each column
                                profiles.profile_header.append(variable)
                                # Set profile database attributes for the esdl asset
                                if not self.io.datetimes[0].tzinfo:
                                    start_date_time = pytz.utc.localize(self.io.datetimes[0])
                                else:
                                    start_date_time = self.io.datetimes[0]
                                if not self.io.datetimes[-1].tzinfo:
                                    end_date_time = pytz.utc.localize(self.io.datetimes[-1])
                                else:
                                    end_date_time = self.io.datetimes[-1]

                                profile_attributes = esdl.InfluxDBProfile(
                                    database=input_energy_system_id,
                                    measurement=asset_name,
                                    field=profiles.profile_header[-1],
                                    port=self.influxdb_port,
                                    host=self.influxdb_host,
                                    startDate=start_date_time,
                                    endDate=end_date_time,
                                    id=str(uuid.uuid4()),
                                )
                                # Assign quantity and units variable
                                if variable in ["Heat_flow"]:
                                    profile_attributes.profileQuantityAndUnit = (
                                        esdl.esdl.QuantityAndUnitType(
                                            physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                            unit=esdl.UnitEnum.WATT,
                                            multiplier=esdl.MultiplierEnum.NONE,
                                        )
                                    )
                                elif variable in [
                                    "HeatIn.H",
                                    "Primary.HeatIn.H",
                                    "Secondary.HeatIn.H",
                                ]:
                                    profile_attributes.profileQuantityAndUnit = (
                                        esdl.esdl.QuantityAndUnitType(
                                            physicalQuantity=esdl.PhysicalQuantityEnum.PRESSURE,
                                            unit=esdl.UnitEnum.PASCAL,
                                            multiplier=esdl.MultiplierEnum.NONE,
                                        )
                                    )
                                elif variable in [
                                    "HeatIn.Q",
                                    "Primary.HeatIn.Q",
                                    "Secondary.HeatIn.Q",
                                ]:
                                    profile_attributes.profileQuantityAndUnit = (
                                        esdl.esdl.QuantityAndUnitType(
                                            physicalQuantity=esdl.PhysicalQuantityEnum.FLOW,
                                            unit=esdl.UnitEnum.CUBIC_METRE,
                                            perTimeUnit=esdl.TimeUnitEnum.SECOND,
                                            multiplier=esdl.MultiplierEnum.NONE,
                                        )
                                    )
                                else:
                                    logger.warning(
                                        f"No profile units will be written to the ESDL for: "
                                        f"{asset_name}. + {variable}"
                                    )

                                asset.port[index_outport].profile.append(profile_attributes)

                            # Add variable values in new column
                            conversion_factor = 0.0
                            if variable in [
                                "HeatIn.H",
                                "Primary.HeatIn.H",
                                "Secondary.HeatIn.H",
                            ]:
                                conversion_factor = GRAVITATIONAL_CONSTANT * 988.0
                            else:
                                conversion_factor = 1.0
                            data_row.append(
                                results[f"{asset_name}." + variable][ii] * conversion_factor
                            )

                        profiles.profile_data_list.append(data_row)
                    # end time steps
                    profiles.num_profile_items = len(profiles.profile_data_list)
                    profiles.start_datetime = profiles.profile_data_list[0][0]
                    profiles.end_datetime = profiles.profile_data_list[-1][0]

                    influxdb_profile_manager = InfluxDBProfileManager(
                        influxdb_conn_settings, profiles
                    )
                    optim_simulation_tag = {"output_esdl_id": energy_system.id}
                    _ = influxdb_profile_manager.save_influxdb(
                        measurement=asset_name,
                        field_names=influxdb_profile_manager.profile_header[1:],
                        tags=optim_simulation_tag,
                    )
                    # -- Test tags -- # do not delete - to be used in test case

                    # prof_loaded_from_influxdb = InfluxDBProfileManager(influxdb_conn_settings)
                    # dicts = [{"tag": "output_esdl_id", "value": energy_system.id}]
                    # prof_loaded_from_influxdb.load_influxdb(
                    #     # '"' + "ResidualHeatSource_72d7" + '"' ,
                    #     '"' + asset_name + '"' ,
                    #     variables_one_hydraulic_system,
                    #     # ["HeatIn.Q"],
                    #     # ["HeatIn.H"],
                    #     # ["Heat_flow"],
                    #     profiles.start_datetime,
                    #     profiles.end_datetime,
                    #     dicts,
                    # )

                    # ------------------------------------------------------------------------------
                    # Do not delete the code below: is used in the development of profile viewer in
                    # mapeditor
                    # Write database to excel file and read in to recreate the database
                    # database name: input esdl id
                    # tags when saving to database: optim_simulation_tag = {"output_esdl_id":
                    # output_esdl_id}

                    # print("Save ESDL profile data to excel")
                    # excel_prof_saved = ExcelProfileManager(
                    #     source_profile=prof_loaded_from_influxdb
                    # )
                    # file_path_setting = (
                    #     f"C:\\Projects_gitlab\\NWN_dev\\rtc-tools-heat-network\\{asset_name}.xlsx"
                    # )
                    # excel_prof_saved.save_excel(
                    #     file_path=file_path_setting,
                    #     sheet_name=input_energy_system_id
                    # )
                    # print("Read data from Excel")
                    # excel_prof_read = ExcelProfileManager()
                    # excel_prof_read.load_excel(file_path_setting)
                    # print("Create database")
                    # influxdb_profile_manager_create_new = InfluxDBProfileManager(
                    #     influxdb_conn_settings, excel_prof_read
                    # )
                    # optim_simulation_tag = {"output_esdl_id": energy_system.id}
                    # _ = influxdb_profile_manager_create_new.save_influxdb(
                    #     measurement=asset_name,
                    #     field_names=influxdb_profile_manager_create_new.profile_header[1:],
                    #     tags=optim_simulation_tag,
                    # )
                    # ------------------------------------------------------------------------------
                except StopIteration:
                    # If the asset has been deleted, thus also not placed
                    pass
                except Exception:  # TODO fix other places in the where try/except end with pass
                    logger.error(
                        f"During the influxDB profile writing for asset: {asset_name}, the "
                        "following error occured:"
                    )
                    traceback.print_exc()
                    sys.exit(1)

            # TODO: create test case
            # Code that can be used to remove a specific measurment from the database
            # try:
            #     influxdb_profile_manager.influxdb_client.drop_measurement(energy_system.id)
            # except:
            #     pass
            # Code that can be used to check if a specific measurement exists in the database
            # influxdb_profile_manager.influxdb_client.get_list_measurements()

            # Do not delete: Test code still to be used in test case
            # try:
            #     esdl_infl_prof = profs[0]
            #     np.any(isinstance(esdl_infl_prof, esdl.InfluxDBProfile))
            # except:
            #     np.any(isinstance(profs, esdl.InfluxDBProfile))
            # print("Reading InfluxDB profile from test...")
            # prof3 = InfluxDBProfileManager(conn_settings)
            # # prof3.load_influxdb("test", ["Heat_flow"])
            # prof3.load_influxdb('"' + energy_system.id + '"', profiles.profile_header[1:4])
            # # can access values via
            # # prof3.profile_data_list[0-row][0/1-date/value],
            # # .strftime("%Y-%m-%dT%H:%M:%SZ")
            # # prof3.profile_data_list[3][0].strftime("%Y-%m-%dT%H:%M:%SZ")
            # ts_prof = prof3.get_esdl_timeseries_profile("Heat_flow")
            # # np.testing.assert_array_equal(ts_prof.values[0], 45)
            # # np.testing.assert_array_equal(ts_prof.values[1], 900)
            # # np.testing.assert_array_equal(ts_prof.values[2], 5.6)
            # # np.testing.assert_array_equal(ts_prof.values[3], 1.2)
            # # np.testing.assert_array_equal(len(ts_prof.values), 4)
            # # -- Test tags --
            # prof3 = InfluxDBProfileManager(influxdb_conn_settings)
            # dicts = [{"tag": "output_esdl_id", "value": energy_system.id}]
            # prof3.load_influxdb(
            #     '"' + "ResidualHeatSource_72d7" + '"' , ["HeatIn.Q"],
            #     profiles.start_datetime,
            #     profiles.end_datetime,
            #     dicts,
            # )
            # test = 0.0
        # ------------------------------------------------------------------------------------------
        # Save esdl file

        self.__optimized_energy_system_handler = esh
        self.optimized_esdl_string = esh.to_string()

        if self.esdl_string is None:
            if optimizer_sim:
                filename = run_info.esdl_file.with_name(
                    f"{run_info.esdl_file.stem}_GrowOptimized.esdl"
                )
            else:
                filename = run_info.esdl_file.with_name(
                    f"{run_info.esdl_file.stem}_Simulation.esdl"
                )
            esh.save(str(filename))



from typing import Dict, List, Tuple, Union

import casadi as ca

import numpy as np

from pymoca.backends.casadi.model import Variable as _Variable


MATHEMATICAL_OPERATORS = [
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    "__mul__",
    "__rmul__",
    "__floordiv__",
    "__truediv__",
    "__mod__",
    "__pow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__xor__",
    "__or__",
    "__iadd__",
    "__isub__",
    "__imul__",
    "__idiv__",
    "__ifloordiv__",
    "__imod__",
    "__ipow__",
    "__ilshift__",
    "__irshift__",
    "__iand__",
    "__ixor__",
    "__ior__",
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    "__complex__",
    "__int__",
    "__long__",
    "__float__",
    "__oct__",
    "__hex__",
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__ge__",
    "__gt__",
]


class BaseVariable(_Variable):
    _attr_set = {"value", "start", "min", "max", "nominal", "fixed"}

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, MATHEMATICAL_OPERATORS[0]):
            for attr in MATHEMATICAL_OPERATORS:

                def _f(self, *args, attr=attr, **kwargs):
                    return getattr(self.symbol, attr)(*args, **kwargs)

                setattr(cls, attr, _f)

        return super().__new__(cls)

    def __init__(self, name, dimensions=None, **kwargs):
        update_attrs = {}

        for k, v in kwargs.items():
            if k in Variable._attr_set:
                if "k" == "start":
                    raise NotImplementedError("Setting 'start' attribute is not supported yet")
                update_attrs[k] = v

        for k in update_attrs:
            kwargs.pop(k)

        if dimensions is None:
            dimensions = []

        super().__init__(ca.MX.sym(name, *dimensions), **kwargs)

        for k, v in update_attrs.items():
            setattr(self, k, v)

    def __MX__(self):  # noqa: N802
        return self.symbol

    def __getitem__(self, key):
        return self.symbol.__getitem__(key)


class Variable(BaseVariable):
    def der(self):
        try:
            return self._derivative
        except AttributeError:
            self._derivative = Variable(f"der({self.symbol.name()})")
            return self._derivative

    @property
    def has_derivative(self):
        return hasattr(self, "_derivative")


class ControlInput(BaseVariable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fixed = False


class ConstantInput(BaseVariable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fixed = True


class SymbolicParameter(BaseVariable):
    def __init__(self, name, *args, **kwargs):
        if args and "value" not in kwargs:
            kwargs["value"] = args[0]
        super().__init__(name, *args, **kwargs)

    @property
    def name(self):
        return self.symbol.name()


class Array:
    def __init__(self, type_, var_name, dimensions, **modifiers):
        self._array = np.empty(dimensions, dtype=object)
        self._names = np.empty(dimensions, dtype=object)

        for index in np.ndindex(dimensions):
            str_suffix = "[{}]".format(",".join(str(x + 1) for x in index))
            self._array[index] = type_(f"{var_name}{str_suffix}", **modifiers)
            self._names[index] = str_suffix

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._array[key - 1]
        else:
            return self._array[tuple(x - 1 for x in key)]


class Model:
    _modifiers = {}
    _variables: Dict[str, Union[Array, BaseVariable, "Model"]] = {}
    _numeric_parameters: Dict[str, Union[float, int, bool, str]] = {}
    _derivatives: Dict[str, Variable] = {}
    _equations: List[ca.MX] = []
    _initial_equations: List[ca.MX] = []
    _inequalities: List[Tuple[ca.MX, float, float]] = []
    _initial_inequalities: List[Tuple[ca.MX, float, float]] = []
    _skip_variables = None

    def __init__(self, name, **modifiers):
        # Note that this method should be such that it's allowed to be called
        # multiple times, e.g. when a (modifiable) parameter is passed as a
        # modifier to its super class.
        self._variables = {}
        self._numeric_parameters = {}
        self._derivatives = {}
        self._equations = []
        self._initial_equations = []

        self.name = name
        # Value assignment can be done directly, but we move it to the value attribute to
        # make sure that all modifiers are a dictionary
        for k, v in modifiers.items():
            if not isinstance(v, dict):
                modifiers[k] = dict(value=v)

        self._modifiers = modifiers

        self.__prefix = "" if name is None else f"{self.name}."

        if Model._skip_variables is None:
            Model._skip_variables = dir(self)

    def add_variable(self, type_, var_name, *dimensions, **kwargs):
        if var_name in self._variables:
            raise Exception(f"Variable with name '{var_name}' already exists")

        if var_name in self._modifiers:
            kwargs = self.merge_modifiers(kwargs, self._modifiers.pop(var_name))
            # Explicit conversion to MX for our wrapper classes
            for k, v in kwargs.items():
                if isinstance(v, BaseVariable):
                    kwargs[k] = ca.MX(v)

        if dimensions:
            var = self._variables[var_name] = Array(
                type_, f"{self.__prefix}{var_name}", dimensions, **kwargs
            )
        else:
            var = self._variables[var_name] = type_(f"{self.__prefix}{var_name}", **kwargs)

        if isinstance(var, (Variable, ControlInput, ConstantInput)) and (
            isinstance(var.value, (ca.MX, BaseVariable)) or not np.isnan(var.value)
        ):
            # For states and algebraic states, we move the "value" part to an equation
            self.add_equation(var - var.value)
            var.value = np.nan

    def add_equation(self, equation, lb=None, ub=None):
        if lb is None and ub is None:
            self._equations.append(equation)
        elif lb is not None and ub is not None and lb == ub:
            self._equations.append(equation - lb)
        else:
            self._inequalities.append((equation, lb, ub))

    def add_initial_equation(self, equation, lb=None, ub=None):
        if lb is None and ub is None:
            self._initial_equations.append(equation)
        elif lb is not None and ub is not None and lb == ub:
            self._initial_equations.append(equation - lb)
        else:
            self._initial_inequalities.append((equation, lb, ub))

    def connect(self, a: "Connector", b: "Connector"):
        if not a.variables.keys() == b.variables.keys():
            raise Exception(
                f"Cannot connect port {a} of type {type(a)} to port {b} "
                f"of type {type(b)} as they have different variables."
            )

        self._equations.extend([a.variables[k] - b.variables[k] for k in a.variables.keys()])

    def der(self, var: Variable):
        return var.der()

    @property
    def variables(self):
        return self._variables.copy()

    @property
    def numeric_parameters(self):
        return self._numeric_parameters.copy()

    @property
    def equations(self):
        return self._equations.copy()

    @property
    def initial_equations(self):
        return self._initial_equations.copy()

    @property
    def inequalities(self):
        return self._inequalities.copy()

    @property
    def initial_inequalities(self):
        return self._initial_inequalities.copy()

    @staticmethod
    def merge_modifiers(a: dict, b: dict):
        """
        Recursive (not in place) merge of dictionaries.

        :param a: Base dictionary to merge.
        :param b: Dictionary to merge on top of base dictionary.
        :return: Merged dictionary
        """
        b = b.copy()

        for k, v in a.items():
            if isinstance(v, dict):
                b_node = b.setdefault(k, {})
                b[k] = Model.merge_modifiers(v, b_node)
            else:
                if k not in b:
                    b[k] = v

        return b

    def __MX__(self):  # noqa: N802
        return self.symbol

    def __getattr__(self, attr):
        try:
            return self._variables[attr]
        except KeyError:
            pass

        try:
            return self._numeric_parameters[attr]
        except KeyError:
            raise AttributeError(f"Attribute '{attr}' not found")

    def __setattr__(self, key, value):
        if self._skip_variables is None or key in self._skip_variables:
            super().__setattr__(key, value)
        else:
            try:
                value = self._modifiers.pop(key)["value"]
            except KeyError:
                pass
            self._numeric_parameters[key] = value

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def flatten(self):
        if self._modifiers:
            raise Exception("Cannot flatten a model with remaining modifiers")

        m = FlattenedModel()

        all_variables = {}
        all_parameters = {}

        all_equations = []
        all_initial_equations = []

        all_inequalities = []
        all_initial_inequalities = []

        # First we expand arrays
        variables = {}
        for k, var in self._variables.items():
            if isinstance(var, Array):
                for el, suff in zip(var._array.ravel(), var._names.ravel()):
                    variables[f"{k}{suff}"] = el
            else:
                variables[k] = var

        # Move variables to flattened model
        for k, var in variables.items():
            if isinstance(var, Model):
                flatten_var = var.flatten()
                all_variables.update(flatten_var._variables)
                all_parameters.update(flatten_var._numeric_parameters)

                all_equations.extend(flatten_var._equations)
                all_initial_equations.extend(flatten_var._initial_equations)

                all_inequalities.extend(flatten_var._inequalities)
                all_initial_inequalities.extend(flatten_var._initial_inequalities)
            else:
                all_variables[f"{self.__prefix}{k}"] = var

        all_parameters.update(
            {f"{self.__prefix}{p}": v for p, v in self._numeric_parameters.items()}
        )
        all_equations.extend(self.equations)
        all_initial_equations.extend(self.initial_equations)

        all_inequalities.extend(self.inequalities)
        all_initial_inequalities.extend(self.initial_inequalities)

        m._variables = all_variables
        m._numeric_parameters = all_parameters

        m._equations = all_equations
        m._initial_equations = all_initial_equations

        m._inequalities = all_inequalities
        m._initial_inequalities = all_initial_inequalities

        return m


class FlattenedModel(Model):
    def __init__(self):
        super().__init__(None)


class Component(Model):
    pass


class Connector(Component):
    pass



from .gas_demand import GasDemand
from .gas_node import GasNode
from .gas_pipe import GasPipe
from .gas_source import GasSource

__all__ = [
    "GasNode",
    "GasPipe",
    "GasSource",
    "GasDemand",
]



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.workflows import NetworkSimulatorHIGHSTestCase

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestNetworkSimulator(TestCase):
    """
    In this test case 2 heat producers and an ATES is used to supply 3 heating demands. A merit
    order (preference of 1st use) is given to the producers: Producer_1 = 2 and Producer_2 = 1.

    Testing:
    - General checks namely demand matching, energy conservation and asset heat variable vs
      calculated heat (based on flow rate)
    - Check that producer 1 (merit oder = 2) is only used for the supply of heat lossed in the
      connected and is does not contribute to the heating demands 1, 2 and 3
    - Check that the ATES is not delivering any heat to the network during the 1st time step
    """

    def test_network_simulator(self):
        import models.test_case_small_network_with_ates.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        solution = run_optimization_problem(NetworkSimulatorHIGHSTestCase, base_folder=base_folder)

        results = solution.extract_results()

        # General checks
        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        # Check that producer 1 (merit oder = 2) is not used
        # and is does not contribute to the heating demands 1, 2 and 3
        np.testing.assert_allclose(
            results["HeatProducer_1.Heat_source"],
            0.0,
            err_msg="Heat producer 1 should be completely disabled "
            ", due to producer 2 being sufficient for "
            "the total heat demand (incl. heat losses) and it has the 1st priority for usage",
            rtol=1.0e-3,
            atol=1.0e-3,
        )
        # Check ATES
        np.testing.assert_array_less(
            results["ATES_033c.Heat_ates"][0],
            0.0,
            err_msg="ATES should not be delivering heat to the network in the 1st time step",
        )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestNetworkSimulator()
    a.test_network_simulator()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class ControlValve(_NonStorageComponent):
    """
    The control valve is a component to create pressure drop. We allow the control valve to create
    pressure drop for flow in both directions. Note that we set the absolute head loss symbol in
    the HeatMixin.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "control_valve"

        self.add_variable(Variable, "dH")

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



import math
from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True)
class PipeClass:
    """
    This class defines a certain pipe. The optimization can select different pipe classes for each
    pipe, depending on the available pipe classes.
    - name: Name of the pipe class e.g. DN200-S1
    - maximum_velocity: Maximum velocity magnitude in the pipe in m/s.
    - u_values: Insulative properties.
    - investment_costs: Cost for the pipe in euro/meter
    """

    name: str
    inner_diameter: float
    maximum_velocity: float
    u_values: Tuple[float, float]
    investment_costs: float

    @property
    def maximum_discharge(self):
        return self.area * self.maximum_velocity

    @property
    def area(self):
        return 0.25 * math.pi * self.inner_diameter**2



from unittest import TestCase

import numpy as np

from rtctools_heat_network._heat_loss_u_values_pipe import heat_loss_u_values_pipe


class TestHeatLossUValues(TestCase):
    def test_scalar_equal_to_single_element_array(self):
        self.assertEqual(
            heat_loss_u_values_pipe(0.15, 0.075, 0.033),
            heat_loss_u_values_pipe(0.15, [0.075], [0.033]),
        )

        self.assertEqual(
            heat_loss_u_values_pipe(0.15, [0.075], [0.033]),
            heat_loss_u_values_pipe(0.15, np.array([0.075]), np.array([0.033])),
        )

    def test_order_of_layers_matters(self):
        self.assertNotEqual(
            heat_loss_u_values_pipe(0.15, [0.075, 0.025], [0.033, 0.25]),
            heat_loss_u_values_pipe(0.15, [0.025, 0.075], [0.25, 0.033]),
        )

    def test_more_insulation_is_less_heat_loss(self):
        # Thicker inner layer
        np.testing.assert_array_less(
            heat_loss_u_values_pipe(0.15, [0.085, 0.025], [0.033, 0.25]),
            heat_loss_u_values_pipe(0.15, [0.075, 0.025], [0.033, 0.25]),
        )

        # Thicker outer layer
        np.testing.assert_array_less(
            heat_loss_u_values_pipe(0.15, [0.075, 0.035], [0.033, 0.25]),
            heat_loss_u_values_pipe(0.15, [0.075, 0.025], [0.033, 0.25]),
        )

        # Lower conductivity
        np.testing.assert_array_less(
            heat_loss_u_values_pipe(0.15, [0.075, 0.025], [0.024, 0.25]),
            heat_loss_u_values_pipe(0.15, [0.075, 0.025], [0.033, 0.25]),
        )

    def test_duplicate_layers_equals_single_layer_of_double_thickness(self):
        np.testing.assert_array_equal(
            heat_loss_u_values_pipe(0.15, [0.05, 0.05], [0.033, 0.033]),
            heat_loss_u_values_pipe(0.15, 0.1, 0.033),
        )



from numpy import nan

from .source import Source


class GeothermalSource(Source):
    """
    The geothermal source component is used to model geothermal doublets. It is equivilent to a
    normal source with the only difference being in the modelling of doublets. The main reason for
    this component instead of using just a regular source is that to have the integer behaviour of
    increasing the amount of doublets. In the HeatMixin an integer is created _aggregation_count to
    model the amount of doublets and the maximum power will scale with this integer instead of
    continuous. This will also ensure that the cost will scale with this integer.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_subtype = "geothermal"

        self.target_flow_rate = nan
        self.single_doublet_power = nan
        self.nr_of_doublets = 1.0



import sys
from pathlib import Path
from typing import Dict
from unittest import TestCase

import numpy as np
import numpy.testing

from rtctools.util import run_optimization_problem

from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.pipe_class import PipeClass


MIP_TOLERANCE = 1e-8


class TestTopoConstraintsOnPipeDiameterSizingExample(TestCase):
    """
    Tests the topo variables and constraints of heat_mixin on the Pipe Diameter Sizing example.
    """

    problem: HeatMixin
    results: Dict[str, np.ndarray]

    @classmethod
    def setUpClass(cls) -> None:
        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
        from examples.pipe_diameter_sizing.src.example import (
            PipeDiameterSizingProblem,
        )  # noqa: E402, I100

        base_folder = (
            Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
        )

        del root_folder
        sys.path.pop(1)

        cls.problem = run_optimization_problem(PipeDiameterSizingProblem, base_folder=base_folder)
        cls.results = cls.problem.extract_results()

    def test_pipe_class_var(self):
        """
        Tests the variables stored in:
        - pipe_topo_pipe_class_var
        - pipe_topo_max_discharge_var
        - pipe_topo_cost_var
        - pipe_diameter_var
        - pipe_heat_loss_var
        """
        for p in self.problem.heat_network_components.get("pipe", []):
            # If there is nothing to choose for the optimizer, no pipe class binaries are made
            if self.problem.pipe_classes(p) is None or len(self.problem.pipe_classes(p)) == 1:
                continue

            class_vars = self.get_pipe_class_vars(p)
            chosen_pc = self.get_chosen_pipe_class(p, class_vars)

            for var_name, value in class_vars.items():
                self.assertTrue(var_name in self.results, msg=f"{var_name} not in results")
                self.assertTrue(
                    abs(value - 0.0) < MIP_TOLERANCE or abs(value - 1.0) < MIP_TOLERANCE,
                    msg=f"Binary {var_name} isn't either 0.0 or 1.0, it is {value}",
                )
            np.testing.assert_almost_equal(
                1.0,
                np.sum(val for val in class_vars.values()),
                err_msg=f"Not exactly 1 pipe class selected for {p}",
            )

            np.testing.assert_array_almost_equal(
                chosen_pc.inner_diameter,
                self.results[f"{p}__hn_diameter"],
                err_msg=f"{p} inner diameter doesn't match expected",
            )
            np.testing.assert_array_almost_equal(
                chosen_pc.investment_costs,
                self.results[f"{p}__hn_cost"],
                err_msg=f"{p} investment costs doesn't match expected",
            )
            np.testing.assert_array_almost_equal(
                chosen_pc.maximum_discharge,
                self.results[f"{p}__hn_max_discharge"],
                err_msg=f"{p} maximum discharge doesn't match expected",
            )

            expected_heat_losses = self.get_heat_losses(p, chosen_pc)
            np.testing.assert_almost_equal(
                self.results[f"{p}__hn_heat_loss"], expected_heat_losses, 5
            )

    def test_pipe_class_ordering_vars(self):
        """
        Tests the variables stored in:
        - pipe_topo_global_pipe_clas_count_var
        - pipe_topo_pipe_class_discharge_ordering_var
        - pipe_topo_pipe_class_cost_ordering_var
        - pipe_topo_pipe_class_heat_loss_ordering_var
        """
        pc_sums = {pc.name: 0 for pc in self.problem.get_unique_pipe_classes()}
        total_pipes_to_optimize = 0
        for p in self.problem.heat_network_components.get("pipe", []):
            # If there is nothing to choose for the optimizer, no pipe class binaries are made,
            if self.problem.pipe_classes(p) is None or len(self.problem.pipe_classes(p)) == 1:
                continue
            pipe_class_vars = self.get_pipe_class_vars(p)
            chosen_pc = self.get_chosen_pipe_class(p, pipe_class_vars)
            pc_sums[chosen_pc.name] += 1
            total_pipes_to_optimize += 1
            for pc in self.problem.pipe_classes(p):
                if p in self.problem.cold_pipes:
                    base_name = f"{self.problem.cold_to_hot_pipe(p)}__hn_pipe_class_{pc.name}"
                else:
                    base_name = f"{p}__hn_pipe_class_{pc.name}"
                cost_ordering_var_name = base_name + "_cost_ordering"
                cost_ordering_var = self.results[cost_ordering_var_name]
                if pc.investment_costs < chosen_pc.investment_costs:
                    np.testing.assert_almost_equal(
                        cost_ordering_var,
                        0.0,
                        err_msg=f"expected the cost order var for {p} and {pc=} to be 0.0, since"
                        f"{chosen_pc=} with higher investment costs",
                    )
                elif pc.investment_costs > chosen_pc.investment_costs:
                    np.testing.assert_almost_equal(
                        cost_ordering_var,
                        1.0,
                        err_msg=f"Expected the cost order var for {p} and {pc=} to be 1.0, since"
                        f"{chosen_pc=} with lower investment costs",
                    )

                discharge_ordering_var_name = base_name + "_discharge_ordering"
                discharge_ordering_var = self.results[discharge_ordering_var_name]
                if pc.maximum_discharge < chosen_pc.maximum_discharge:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        0.0,
                        err_msg=f"expected the discharge order var for {p} and {pc=} to be 0.0, "
                        f"since {chosen_pc=} with higher max discharge",
                    )
                elif pc.maximum_discharge > chosen_pc.maximum_discharge:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        1.0,
                        err_msg=f"Expected the discharge order var for {p} and {pc=} to be 1.0, "
                        f"since {chosen_pc=} with lower max discharge",
                    )

                heat_loss_ordering_var_name = base_name + "_heat_loss_ordering"
                heat_loss_ordering_var = self.results[heat_loss_ordering_var_name]
                pc_heat_loss = self.get_heat_losses(p, pc)
                chosen_pc_heat_loss = self.get_heat_losses(p, chosen_pc)
                if pc_heat_loss < chosen_pc_heat_loss:
                    np.testing.assert_almost_equal(
                        heat_loss_ordering_var,
                        0.0,
                        err_msg=f"expected the heat loss order var for {p} and {pc=} to be 0.0, "
                        f"since {chosen_pc=} with higher heat losses",
                    )
                elif pc_heat_loss > chosen_pc_heat_loss:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        1.0,
                        err_msg=f"Expected the heat loss order var for {p} and {pc=} to be 1.0, "
                        f"since {chosen_pc=} with lower heat losses",
                    )

        for pc_name, total_count in pc_sums.items():
            count_var_value = self.results[f"{pc_name}__global_pipe_class_count"]
            np.testing.assert_almost_equal(
                count_var_value,
                total_count,
                err_msg=f"Pipe count for {pc_name} doesn't match the expected {total_count}",
            )
        total_pipe_count = sum(pc_sums.values())
        np.testing.assert_equal(
            total_pipe_count,
            total_pipes_to_optimize,
            err_msg=f"Found {total_pipe_count} total selected pipe classes, "
            f"but {total_pipes_to_optimize=}",
        )

    def get_pipe_class_vars(self, pipe: str) -> Dict[str, np.ndarray]:
        given_pipe_classes = self.problem.pipe_classes(pipe)
        if pipe in self.problem.cold_pipes:
            expected_class_vars = [
                f"{self.problem.cold_to_hot_pipe(pipe)}__hn_pipe_class_{pc.name}"
                for pc in given_pipe_classes
            ]
        else:
            expected_class_vars = [f"{pipe}__hn_pipe_class_{pc.name}" for pc in given_pipe_classes]
        class_vars = {
            var_name: value
            for var_name, value in self.results.items()
            if var_name in expected_class_vars
        }
        return class_vars

    def get_chosen_pipe_class(self, pipe: str, pipe_class_vars: Dict[str, np.ndarray]) -> PipeClass:
        chosen_var = None
        given_pipe_classes = self.problem.pipe_classes(pipe)
        for var_name, value in pipe_class_vars.items():
            if abs(value - 1.0) < MIP_TOLERANCE:
                chosen_var = var_name
        self.assertIsNotNone(chosen_var, msg=f"No pipe class selected for {pipe}")
        class_name = chosen_var.split("_")[-1]
        chosen_pc = [pc for pc in given_pipe_classes if pc.name == class_name]
        self.assertEqual(len(chosen_pc), 1, msg=f"Found multiple chosen pipe classes for {pipe}")
        return chosen_pc[0]

    def get_heat_losses(self, pipe: str, pipe_class: PipeClass):
        return self.problem._pipe_heat_loss(
            options=self.problem.heat_network_options(),
            parameters=self.problem.parameters(0),
            p=pipe,
            u_values=pipe_class.u_values,
        )



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestESDL(TestCase):
    def test_basic_source_and_demand_heat(self):
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatESDL, HeatPython

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent

        case_python = run_optimization_problem(HeatPython, base_folder=base_folder)
        case_esdl = run_optimization_problem(HeatESDL, base_folder=base_folder)

        self.assertAlmostEqual(case_python.objective_value, case_esdl.objective_value, 6)

        np.testing.assert_allclose(
            case_python.extract_results()["demand.Heat_demand"],
            case_esdl.extract_results()["demand.Heat_demand"],
        )

    # def test_basic_source_and_demand_qth(self):
    #     import models.basic_source_and_demand.src.qth_comparison as qth_comparison
    #     from models.basic_source_and_demand.src.qth_comparison import QTHPython, QTHESDL
    #
    #     base_folder = Path(qth_comparison.__file__).resolve().parent.parent
    #
    #     case_python = run_optimization_problem(QTHPython, base_folder=base_folder)
    #     case_esdl = run_optimization_problem(QTHESDL, base_folder=base_folder)
    #
    #     np.testing.assert_allclose(
    #         case_python._objective_values[0], case_esdl._objective_values, rtol=1e-5, atol=1e-5
    #     )



from ._internal import HeatComponent
from .heat_port import HeatPort


class HeatTwoPort(HeatComponent):
    """
    The HeatTwoPort component is used as a base for interaction with one hydraulically coupled
    system. As heat networks are closed systems we always need two ports to model both the in and
    out going flow in the system.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(HeatPort, "HeatIn")
        self.add_variable(HeatPort, "HeatOut")



from numpy import nan

from rtctools_heat_network.pycml import Variable
from rtctools_heat_network.pycml.component_library.heat._internal import BaseAsset
from rtctools_heat_network.pycml.component_library.heat.heat_four_port import HeatFourPort


class HeatPump(HeatFourPort, BaseAsset):
    """
    The heatpump component is used to model a water-water heatpump.
    A constant COP is used to model the electricity use of the heatpump. A power cap is set
    on the primary side to model physical constraints on the amount of heat transfer.

    The heat to discharge constraints are set in the HeatMixin. The primary side is modelled as a
    demand, meaning it consumes energy from the primary network and gives it to the secondary side,
    where the secondary side acts like a source to the secondary network. This also means that heat
    can only flow from primary to secondary.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(),
                modifiers,
            ),
        )

        self.component_type = "heat_pump"
        self.COP = nan  # TODO: maybe set this to a standard value if not set in esdl.
        self.nominal = (
            self.Secondary.Q_nominal * self.Secondary.rho * self.Secondary.cp * self.Secondary.dT
        )
        self.elec_power_nominal = self.nominal / self.COP

        # Assumption: heat in/out and added is nonnegative

        self.add_variable(Variable, "Primary_heat", min=0.0)
        self.add_variable(Variable, "Secondary_heat", min=0.0)
        self.add_variable(Variable, "Heat_flow", nominal=self.nominal)
        self.add_variable(Variable, "Power_elec", min=0.0)
        self.add_variable(Variable, "dH_prim")
        self.add_variable(Variable, "dH_sec")

        # Hydraulically decoupled so Heads remain the same
        # #TODO: can't these two equations be moved to the non_storagecomponent?
        self.add_equation(self.dH_prim - (self.Primary.HeatOut.H - self.Primary.HeatIn.H))
        self.add_equation(self.dH_sec - (self.Secondary.HeatOut.H - self.Secondary.HeatIn.H))

        self.add_equation(
            ((self.Primary_heat + self.Power_elec - self.Secondary_heat) / self.nominal)
        )

        self.add_equation(((self.Secondary_heat - self.COP * self.Power_elec) / self.nominal))

        self.add_equation(
            (
                (self.Primary_heat - (self.Primary.HeatIn.Heat - self.Primary.HeatOut.Heat))
                / self.nominal
            )
        )
        self.add_equation(
            (
                (self.Secondary_heat - (self.Secondary.HeatOut.Heat - self.Secondary.HeatIn.Heat))
                / self.nominal
            )
        )
        self.add_equation((self.Heat_flow - self.Secondary_heat) / self.nominal)



from rtctools_heat_network.pycml import Variable

from ._internal import BaseAsset, HeatComponent
from .heat_port import HeatPort


class Node(HeatComponent, BaseAsset):
    """
    A node is the only component in the network that allows to model 3 or more flows to come
    together. This essentially means that only pipes can be connected to ports and that it models
    junctions where multiple pipes come together. The node ensures that the heat on all ports is
    equal. Furthermore, it ensures that discharge and heat are conserved for which constraints in
    the HeatMixin are set.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "node"

        self.n = 2
        assert self.n >= 2

        self.add_variable(HeatPort, "HeatConn", self.n)
        self.add_variable(Variable, "H")

        # Because the orientation of the connected pipes are important to
        # setup the heat conservation, these constraints are added in the
        # mixin.

        for i in range(1, self.n + 1):
            self.add_equation(self.HeatConn[i].H - self.H)
            # Q and Heat to be set in the mixin



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class Demand(_NonStorageComponent):
    """
    The demand component is there to extract thermal power (Heat) out of the network. Typically,
    this component is used to model aggregated demands.

    The heat to discharge constraints are set in the HeatMixin. We enforce that the outgoing
    temperature of the demand matches the absolute thermal power, Q * cp * rho * T_ret == Heat,
    similar as with the sources. This allows us to guarantee that the flow can always carry the
    heat and that thermal losses are always estimated conservatively, as the heat losses further
    downstream in the network are over-estimated with T_ret where in reality this temperature
    drops. It also implicitly assumes that the temperature drops in the network are small and thus
    satisfy minimum temperature requirements.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    HeatIn=dict(Heat=dict(min=0.0)),
                    HeatOut=dict(Heat=dict(min=0.0)),
                ),
                modifiers,
            ),
        )

        self.component_type = "demand"

        # Assumption: heat in/out and extracted is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(Variable, "Heat_demand", min=0.0, nominal=self.Heat_nominal)

        self.add_equation(
            (self.HeatOut.Heat - (self.HeatIn.Heat - self.Heat_demand)) / self.Heat_nominal
        )

        self.add_equation((self.Heat_flow - self.Heat_demand) / self.Heat_nominal)



from rtctools_heat_network.pycml import Variable

from .gas_base import GasPort
from .._internal import BaseAsset
from .._internal.gas_component import GasComponent


class GasDemand(GasComponent, BaseAsset):
    """
    A gas demand consumes flow from the network.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_demand"
        self.min_head = 30.0

        self.add_variable(GasPort, "GasIn")
        self.add_variable(Variable, "Gas_demand_flow", min=0.0)

        self.add_equation((self.GasIn.Q - self.Gas_demand_flow))
